def register_queue_handler(clazz):
    from rapid.lib.queue_handler_constants import QueueHandlerConstants
    if clazz not in QueueHandlerConstants.queue_handler_classes:
        QueueHandlerConstants.queue_handler_classes.append(clazz)
    return clazz



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
import os
import time
import pickle
import socket
import logging
import requests
from requests.exceptions import ConnectionError  # pylint: disable=redefined-builtin

from rapid.client.client_configuration import ClientConfiguration
from rapid.lib.constants import HeaderConstants
from rapid.lib.utils import OSUtil

from rapid.lib.version import Version

try:
    import simplejson as json
except ImportError:
    import json

from rapid.lib.communicator import Communicator
from rapid.lib.store_service import StoreService

logger = logging.getLogger("rapid")

# pylint: disable=broad-except


class ClientCommunicator(Communicator):
    DOWNLOAD_URI = '/get_file/{}'
    REGISTRATION_URI = "/client/register"
    WORK_REQUEST_URI = "/api/action_instances/{}/work_request"

    def __init__(self, server_uri, quarantine_directory=None, flask_app=None, verify_certs=True, get_files_auth=None):
        """

        :param server_uri:
        :param quarantine_directory:
        :param flask_app:
        :param verify_certs:
        :type get_files_auth: :class:`requests.auth.HTTPBasicAuth`
        :return:
        """
        super(ClientCommunicator, self).__init__(server_uri, flask_app=flask_app)
        self.quarantine_directory = quarantine_directory
        self.verify_certs = verify_certs if not flask_app else flask_app.rapid_config.verify_certs
        self.get_files_auth = get_files_auth

    def get_work_request_by_action_instance_id(self, action_instance_id):
        return self._default_get(self.get_uri(self.WORK_REQUEST_URI.format(action_instance_id))).json()

    def get_file(self, directory, file_name, logger_in):
        """
        Download via requests, the file that is requested.
        :param logger_in:
        :param file_name:
        :return:
        """
        file_name = file_name.strip()
        uri = self.get_uri(self.DOWNLOAD_URI.format(file_name))
        try:
            local_file_name = self.get_downloaded_file_name(directory, file_name)
            return local_file_name
        except Exception:
            import traceback
            traceback.print_exc()
            logger_in.error("unable to download file: {}".format(uri))

    def send_parameters(self, pipeline_instance_id, parameters, logger_in):
        if parameters:
            try:
                self._default_send(self.api_uri("pipeline_instance/{}".format(pipeline_instance_id)), parameters)
            except Exception as exception:
                logger_in.error(exception)

    def send_statistics(self, action_instance_id, stats, logger_in):
        if stats:
            try:
                self._default_send(self.api_uri("action_instance/{}/stats".format(action_instance_id)), stats)
            except Exception as exception:
                logger_in.error(exception)

    def send_results(self, action_instance_id, results, logger_in):
        if results:
            try:
                self._default_send(self.api_uri("action_instance/{}/results".format(action_instance_id)), results)
            except Exception as exception:
                logger_in.error(exception)

    def send_test_analysis(self, pipeline_instance_id, analysis, logger_in):
        if analysis:
            try:
                self._default_send(self.api_uri("qa_tests/analysis"), None, 'post', {'X-Pipeline-Instance-Id': str(pipeline_instance_id)}, in_json=analysis)
            except Exception as exception:
                logger_in.error(exception)

    def send_done(self, action_instance_id, status, parameters, stats, results, logger_in, headers=None):
        data = {'status': status}

        if headers is None:
            headers = {"Content-Type": 'application/json'}
        else:
            headers['Content-Type'] = 'application/json'

        if parameters:
            data['parameters'] = parameters
        if stats:
            data['stats'] = stats
        if results:
            data['results'] = results

        try:
            self._default_send(self.api_uri("action_instances/{}/done".format(action_instance_id)), json.dumps(data), 'post',
                               headers)
        except Exception as exception:
            logger_in.error(exception)
            self._handle_quarantine_failure(exception, action_instance_id, data, logger_in)

    def _handle_quarantine_failure(self, exception, action_instance_id, data, logger_in):
        if self.quarantine_directory is not None \
                and ((isinstance(exception, ConnectionError) and ("Connection refused" in "{}".format(exception) or
                                                                  "Connection aborted" in "{}".format(exception))) or
                     'Status Code Failure:' in exception.message):
            try:
                os.makedirs(self.quarantine_directory)
            except Exception:
                pass

            message = "Master not available, storing for later: {}".format(action_instance_id)
            logger_in.error(message)

            with open('{}/{}'.format(self.quarantine_directory, action_instance_id), 'w') as tmp_file:
                tmp_file.writelines(pickle.dumps(data))
                tmp_file.flush()

    @staticmethod
    def _get_register_post_data(client_config):
        return json.dumps({"grains": client_config.grains,
                           "grain_restrict": client_config.grain_restrict,
                           "hostname": socket.gethostname()})

    @staticmethod
    def _get_register_headers(client_config):
        headers = {'Content-Type': 'application/json',
                   'X-RAPIDCI-PORT': str(client_config.port) if hasattr(client_config, 'port') else '',
                   'X-RAPIDCI-REGISTER-KEY': client_config.register_api_key if hasattr(client_config, 'register_api_key') else '',
                   'X-RAPIDCI-TIME': str(time.time() * 1000),
                   'X-RAPIDCI-CLIENT-KEY': client_config.api_key if hasattr(client_config, 'api_key') else ''}
        if hasattr(client_config, 'use_ssl') and client_config.use_ssl:
            headers['X-Is-Ssl'] = 'true'

        try:
            if client_config.is_single_use:
                headers[HeaderConstants.SINGLE_USE] = 'true'
        except (AttributeError, TypeError):
            pass

        return headers

    def register(self, client_config):
        # type: (ClientConfiguration) -> dict
        try:
            response = self._default_send(self.get_uri(self.REGISTRATION_URI),
                                          self._get_register_post_data(client_config),
                                          'post',
                                          self._get_register_headers(client_config))

            if response.status_code == 200 and 'X-Rapidci-Master-Key' in response.headers:
                StoreService.save_master_key(self.flask_app, response.headers['X-Rapidci-Master-Key'])
            return {'master_version': response.headers[Version.HEADER]}
        except Exception as exception:
            logger.error(exception)

    def _string_header_values(self, headers):
        _headers = {}
        try:
            for key, value in headers.items():
                _headers[key] = str(value) if value is not None else ''
        except (TypeError, AttributeError):
            pass
        return _headers

    def _default_get(self, uri):
        headers = {}
        master_key = StoreService.get_master_key(self.flask_app)
        headers = self._string_header_values(headers)

        if master_key is not None:
            headers['X-Rapidci-Api-Key'] = master_key

        return requests.get(uri, headers=headers)

    def _default_send(self, uri, data, in_type='put', headers=None, in_json=None):
        master_key = StoreService.get_master_key(self.flask_app)
        headers = self._string_header_values(headers)
        
        if master_key is not None:
            headers['X-Rapidci-Api-Key'] = master_key

        request = None
        if in_type == 'put':
            request = requests.put(uri, data=data, json=in_json, headers=headers, verify=self.verify_certs)
        elif in_type == 'post':
            request = requests.post(uri, data=data, json=in_json, headers=headers, verify=self.verify_certs)

        if request.status_code != 200:
            raise Exception("Status Code Failure: {}".format(request.status_code))
        return request

    def _get_real_file_name(self, directory, file_name):
        return OSUtil.path_join(directory, file_name.split('/')[-1])  # required files must ALWAYS be with / not \

    def get_downloaded_file_name(self, directory, file_name):
        real_file_name = self._get_real_file_name(directory, file_name)
        try:
            os.makedirs(os.path.dirname(real_file_name))
        except Exception:
            pass

        with open(real_file_name, 'wb') as handle:
            response = None
            if self.get_files_auth is not None:
                response = requests.get(self.server_uri + "/get_file/{}".format(file_name), verify=self.verify_certs, auth=self.get_files_auth)
            else:
                response = requests.get(self.server_uri + "/get_file/{}".format(file_name), verify=self.verify_certs)

            if not response.ok:
                raise BaseException("File '{}' did not download".format(file_name))

            for block in response.iter_content(1024):
                handle.write(block)
        return real_file_name



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

status_summary

Revision ID: a08ae00f2ba7
Revises: e16dacfc467e
Create Date: 2016-02-03 17:56:13.138231

"""

# revision identifiers, used by Alembic.
revision = 'a08ae00f2ba7'
down_revision = 'e16dacfc467e'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.create_table('qa_status_summaries',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('action_instance_id', sa.Integer(), nullable=False),
    sa.Column('status_id', sa.Integer(), nullable=False),
    sa.Column('count', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['action_instance_id'], ['action_instances.id'], ),
    sa.ForeignKeyConstraint(['status_id'], ['statuses.id'], ),
    sa.PrimaryKeyConstraint('id'),
    mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_qa_status_summaries_action_instance_id'), 'qa_status_summaries', ['action_instance_id'], unique=False)
    op.create_index(op.f('ix_qa_status_summaries_id'), 'qa_status_summaries', ['id'], unique=False)
    op.create_index(op.f('ix_qa_status_summaries_status_id'), 'qa_status_summaries', ['status_id'], unique=False)
    op.create_index(op.f('ix_commit_parameters_commit_id'), 'commit_parameters', ['commit_id'], unique=False)
    op.create_index(op.f('ix_commit_parameters_id'), 'commit_parameters', ['id'], unique=False)
    op.create_index(op.f('ix_commit_parameters_name'), 'commit_parameters', ['name'], unique=False)
    op.drop_index('ix_commit_param_commit_id', table_name='commit_parameters')
    op.drop_index('ix_commit_param_id', table_name='commit_parameters')
    op.drop_index('ix_commit_param_name', table_name='commit_parameters')
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.create_index('ix_commit_param_name', 'commit_parameters', ['name'], unique=False)
    op.create_index('ix_commit_param_id', 'commit_parameters', ['id'], unique=False)
    op.create_index('ix_commit_param_commit_id', 'commit_parameters', ['commit_id'], unique=False)
    op.drop_index(op.f('ix_commit_parameters_name'), table_name='commit_parameters')
    op.drop_index(op.f('ix_commit_parameters_id'), table_name='commit_parameters')
    op.drop_index(op.f('ix_commit_parameters_commit_id'), table_name='commit_parameters')
    op.drop_index(op.f('ix_qa_status_summaries_status_id'), table_name='qa_status_summaries')
    op.drop_index(op.f('ix_qa_status_summaries_id'), table_name='qa_status_summaries')
    op.drop_index(op.f('ix_qa_status_summaries_action_instance_id'), table_name='qa_status_summaries')
    op.drop_table('qa_status_summaries')
    ### end Alembic commands ###



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
# pylint: disable=no-member,too-few-public-methods,broad-except
import datetime

from sqlalchemy.orm import relationship
from sqlalchemy import Column, String, ForeignKey, Integer, Boolean, Text

from rapid.lib import get_declarative_base
from rapid.lib.constants import StatusConstants
from rapid.lib.converters import ObjectConverter
from rapid.master.data.database.models.base.base_model import BaseModel
from rapid.master.data.database.models.base.date_model import DateModel
from rapid.master.data.database.models.base.active_model import ActiveModel
Base = get_declarative_base()


class Action(BaseModel, Base):
    cmd = Column(String(255), nullable=False)
    name = Column(String(100), nullable=False)
    executable = Column(String(100), nullable=False)
    args = Column(String(255), nullable=False)
    order = Column(Integer, nullable=False, index=True)
    manual = Column(Boolean, default=False, nullable=False)
    callback_required = Column(Boolean, default=False, nullable=False)
    grain = Column(String(100))
    slices = Column(Integer, default=0)

    workflow_id = Column(Integer, ForeignKey('workflows.id'), nullable=False, index=True)
    pipeline_id = Column(Integer, ForeignKey("pipelines.id"), nullable=False, index=True)

    configuration = relationship('ActionConfig', lazy='select', uselist=False, backref="parent")

    def convert_to_instance(self):
        action_instance = ActionInstance()
        action_instance.status_id = StatusConstants.NEW
        action_instance.action_id = self.id

        return ObjectConverter.copy_attributes(self, action_instance, ['cmd',
                                                                       'executable',
                                                                       'args',
                                                                       'order',
                                                                       'manual',
                                                                       'callback_required',
                                                                       'grain'])


class ActionConfig(BaseModel, Base):
    action_id = Column(Integer, ForeignKey('actions.id'), nullable=False, index=True)
    configuration = Column(Text, nullable=False)


class ActionInstance(DateModel, BaseModel, Base):
    cmd = Column(String(255), nullable=False)
    executable = Column(String(100), nullable=False)
    args = Column(String(255), nullable=False)
    order = Column(Integer, nullable=False, default=0, index=True)
    manual = Column(Boolean, nullable=False, default=False)
    callback_required = Column(Boolean, nullable=False, default=False)
    grain = Column(String(100))
    assigned_to = Column(String(150), nullable=True)
    slice = Column(String(25), default='')

    status_id = Column(Integer, ForeignKey('statuses.id'), nullable=False, index=True)
    action_id = Column(Integer, ForeignKey('actions.id'), nullable=False, index=True)

    workflow_instance_id = Column(Integer, ForeignKey('workflow_instances.id'), nullable=False, index=True)
    pipeline_instance_id = Column(Integer, ForeignKey("pipeline_instances.id"), nullable=False, index=True)

    configuration = relationship('ActionInstanceConfig', lazy='select', uselist=False, backref="parent")
    status = relationship('Status')
    pipeline_instance = relationship('PipelineInstance')
    action = relationship('Action')


class ActionInstanceConfig(BaseModel, Base):
    action_instance_id = Column(Integer, ForeignKey('action_instances.id'), nullable=False, index=True)
    configuration = Column(Text, nullable=False)


class AppConfiguration(BaseModel, Base):
    process_queue = Column(Boolean, default=True)


class IntegrationType(BaseModel, Base):
    name = Column(String(255), nullable=False)
    active = Column(Boolean, default=True, nullable=False)


class IntegrationKeys(BaseModel, Base):
    name = Column(String(255), nullable=False)
    active = Column(Boolean, default=True, nullable=False)


class Integration(BaseModel, Base):
    integration_type_id = Column(Integer, ForeignKey('integration_types.id'), nullable=False)
    integration_keys_id = Column(Integer, ForeignKey('integration_keys.id'), nullable=False)
    value = Column(String(500), nullable=False)


class Pipeline(ActiveModel, Base):
    stages = relationship("Stage", backref="pipeline", order_by="asc(Stage.order)")

    pipeline_events = relationship('PipelineEvent', backref="pipeline")

    def convert_to_instance(self):
        instance = PipelineInstance()
        instance.pipeline_id = self.id                                                                                                                                   
        return ObjectConverter.copy_attributes(self, instance, [])


class PipelineInstance(BaseModel, DateModel, Base):
    pipeline_id = Column(Integer, ForeignKey('pipelines.id'), nullable=False, index=True)
    status_id = Column(Integer, ForeignKey('statuses.id'), nullable=False, index=True)
    priority = Column(Integer, default=0, index=True)

    stage_instances = relationship("StageInstance", backref="pipeline_instance", order_by="asc(StageInstance.order)")
    action_instances = relationship("ActionInstance", order_by="ActionInstance.id")
    parameters = relationship('PipelineParameters', backref="pipeline_instance")
    stats = relationship('PipelineStatistics', backref="pipeline_instance")
    pipeline = relationship('Pipeline')
    status = relationship('Status')

    def get_parameters_dict(self):
        results = {}
        try:
            for parameter in self.parameters:
                results[parameter.parameter] = parameter.value

            results['pipeline_instance_id'] = self.id
        except Exception:
            pass
        return results


class PipelineParameters(BaseModel, Base):
    pipeline_instance_id = Column(Integer, ForeignKey("pipeline_instances.id"), nullable=False, index=True)
    parameter = Column(String(200), nullable=False, index=True)
    value = Column(String(4000), nullable=False)


class PipelineStatistics(BaseModel, Base):
    pipeline_instance_id = Column(Integer, ForeignKey("pipeline_instances.id"), nullable=False, index=True)
    statistics_id = Column(Integer, ForeignKey("statistics.id"), nullable=False, index=True)
    value = Column(Integer, nullable=False)

    statistics = relationship('Statistics', backref="pipeline_statistics")


class Stage(ActiveModel, Base):
    order = Column(Integer, nullable=False, default=0, index=True)
    pipeline_id = Column(Integer, ForeignKey("pipelines.id"), nullable=False, index=True)

    workflows = relationship("Workflow", backref="stage")

    def convert_to_instance(self):
        instance = StageInstance()
        instance.status_id = StatusConstants.NEW
        instance.stage_id = self.id
        instance.start_date = datetime.datetime.utcnow()
        return ObjectConverter.copy_attributes(self, instance, ['order'])


class StageInstance(BaseModel, DateModel, Base):
    order = Column(Integer, nullable=False, default=0, index=True)
    status_id = Column(Integer, ForeignKey("statuses.id"), nullable=False, index=True)
    pipeline_instance_id = Column(Integer, ForeignKey("pipeline_instances.id"), nullable=False, index=True)
    stage_id = Column(Integer, ForeignKey("stages.id"), nullable=False, index=True)

    stage = relationship('Stage')
    status = relationship('Status')

    workflow_instances = relationship("WorkflowInstance", backref="stage_instance", order_by="asc(WorkflowInstance.order)")


class Statistics(ActiveModel, Base):
    pass


class Status(ActiveModel, Base):
    type = Column(String(50), nullable=False, index=True)
    display_name = Column(String(100), nullable=False)


class Workflow(ActiveModel, Base):
    order = Column(Integer, nullable=False, default=0, index=True)
    stage_id = Column(Integer, ForeignKey("stages.id"), nullable=False, index=True)

    actions = relationship("Action", backref="workflow", order_by="asc(Action.order)")

    def convert_to_instance(self):
        instance = WorkflowInstance()
        instance.status_id = StatusConstants.NEW
        instance.workflow_id = self.id
        instance.start_date = datetime.datetime.utcnow()
        return ObjectConverter.copy_attributes(self, instance, ['order'])


class WorkflowInstance(BaseModel, DateModel, Base):
    order = Column(Integer, nullable=False, default=0, index=True)
    status_id = Column(Integer, ForeignKey("statuses.id"), nullable=False, index=True)
    stage_instance_id = Column(Integer, ForeignKey("stage_instances.id"), nullable=False, index=True)
    workflow_id = Column(Integer, ForeignKey('workflows.id'), nullable=False, index=True)

    status = relationship('Status')
    workflow = relationship('Workflow')
    action_instances = relationship('ActionInstance', backref="workflow_instance", order_by="asc(ActionInstance.order)")


class EventType(ActiveModel, Base):
    pass


class PipelineEvent(BaseModel, Base):
    pipeline_id = Column(ForeignKey('pipelines.id'), nullable=False, index=True)
    event_type_id = Column(ForeignKey('event_types.id'), nullable=False, index=True)
    conditional = Column(Text, nullable=False)
    config = Column(Text, nullable=False)





import logging
import importlib

from rapid.lib.configuration import Configuration
from rapid.lib.framework.injectable import Injectable
logger = logging.getLogger('rapid')


class ExtensionLoader(Injectable):
    __injectables__ = {'rapid_config': None}

    def __init__(self, rapid_config):
        self.rapid_config = rapid_config

    def load_extensions(self, flask_app):
        if self.rapid_config.extensions:
            for extension in self.rapid_config.extensions:
                try:
                    extension_module = importlib.import_module(extension)
                    if hasattr(extension_module, 'setup_rapid'):
                        extension_module.setup_rapid(flask_app, self.rapid_config)
                        logger.info('The extension "{}" was loaded.'.format(extension))
                        continue
                    logger.info("The extension '{}' does not have a 'setup_rapid' method.".format(extension))
                except (ModuleNotFoundError, ImportError):
                    logger.info('The extension named "{}" was not loaded.'.format(extension))



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

import socket
from abc import ABCMeta, abstractmethod
import os
try:
    from ConfigParser import SafeConfigParser as ConfigParser
except ImportError:
    from configparser import ConfigParser

import logging

logger = logging.getLogger("rapid")


class ConfigurationDefaults(object):
    def __init__(self, default=None, type_cast=str, delim=','):
        self._default = default
        self._type_cast = type_cast
        self._delim = delim

    @property
    def defaults(self):
        return [self._default, self._type_cast, self._delim]


class Configuration(object):
    __metaclass__ = ABCMeta

    def __init__(self, file_name=None):
        self.hostname = socket.gethostname()
        self.extensions = None

        self.parse_config_file(file_name)

    def parse_config_file(self, file_name):
        parser = ConfigParser()
        try:
            if file_name:
                if os.path.exists(file_name):
                    parser.read(file_name)
        except Exception:  # pylint: disable=bare-except
            logger.info("Using defaults")

        self._set_values(parser)
        return self

    @property
    @abstractmethod
    def section_mapping(self):
        """
        Returns
        -------
        dict
            dict of key, dict of key value mappings.
        """
        yield

    def get_section(self, key):
        for section, value in self.section_mapping.items():
            if key in value:
                return section
        return None

    def _set_values(self, parser):
        for section, map in self.section_mapping.items():
            for key, value in map.items():
                try:
                    self._set_parser_value(parser, section, key, *value.defaults)
                except AttributeError:
                    self._set_parser_value(parser, section, key, *value)

    def _handle_normal_value(self, parser, key, section, type_cast):
        setattr(self, key, type_cast(parser.get(section, key)))

    def _set_parser_value(self, parser, section, key, default=None, type_cast=str, delim=','):
        try:
            if type_cast == bool:
                setattr(self, key, parser.get(section, key).lower().strip() == 'true')
            elif type_cast == list:
                setattr(self, key, parser.get(section, key).split(delim))
            else:
                self._handle_normal_value(parser, key, section, type_cast)
        except Exception:  # pylint: disable=broad-except
            setattr(self, key, default)



"""
Copyright (c) 2015 Michael Bright and Bamboo HR LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

vcs_releases

Revision ID: 28fd8ab077e8
Revises: ce9be6e8354c
Create Date: 2018-06-05 18:46:26.844276

"""

# revision identifiers, used by Alembic.
revision = '28fd8ab077e8'
down_revision = 'ce9be6e8354c'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
from rapid.lib.constants import VcsReleaseStepType

def upgrade():
    op.create_table('vcs_releases',
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('search_filter', sa.String(length=500), nullable=False),
                    sa.Column('notification_id', sa.String(length=250), nullable=False),
                    sa.Column('vcs_id', sa.Integer(), nullable=False),
                    sa.ForeignKeyConstraint(['vcs_id'], ['vcs.id'], ),
                    sa.PrimaryKeyConstraint('id'),
                    mysql_engine='InnoDB'
                    )
    op.create_index(op.f('ix_vcs_releases_id'), 'vcs_releases', ['id'], unique=False)
    op.create_index(op.f('ix_vcs_releases_vcs_id'), 'vcs_releases', ['vcs_id'], unique=False)

    op.create_table('vcs_release_steps',
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('name', sa.String(length=250), nullable=False),
                    sa.Column('custom_id', sa.String(length=250), nullable=False),
                    sa.Column('user_required', sa.Boolean(), nullable=False),
                    sa.Column('sort_order', sa.Integer(), nullable=True),
                    sa.Column('type', sa.Enum(*list(map(lambda x: x.name, VcsReleaseStepType))), nullable=False),
                    sa.Column('vcs_release_id', sa.Integer(), nullable=False),
                    sa.ForeignKeyConstraint(['vcs_release_id'], ['vcs_releases.id'], ),
                    sa.PrimaryKeyConstraint('id'),
                    mysql_engine='InnoDB'
                    )
    op.create_index(op.f('ix_vcs_release_steps_id'), 'vcs_release_steps', ['id'], unique=False)
    op.create_index(op.f('ix_vcs_release_steps_vcs_release_id'), 'vcs_release_steps', ['vcs_release_id'], unique=False)


def downgrade():
    op.drop_table('vcs_release_steps')
    op.drop_table('vcs_releases')



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
from rapid.lib.constants import Constants

parsers = {}


def register(clz):
    parsers[clz.get_type()] = clz
    return clz


def load_parsers():
    from .xunit_parser import XUnitParser
    from .tap_parser import TapParser

    for parser in [XUnitParser, TapParser]:
        register(parser)


def parse_file(lines, parser=None, workspace=None):
    if parser is None and lines[0] and lines[0].strip() in parsers:
        try:
            parser = parsers[lines[0].strip()](workspace=workspace)
            return parser.parse(lines)
        except Exception as exception:
            import traceback
            traceback.print_exc()
            raise exception
    elif parser:
        try:
            return parser.parse(lines, ignore_type_check=True)
        except Exception as exception:
            import traceback
            traceback.print_exc()
            raise exception
    return {}


def get_parser(identifier, workspace=None, failures_only=False, failures_count=False):
    if identifier in parsers:
        return parsers[identifier](workspace=workspace, failures_only=failures_only, failures_count=failures_count)
    return None


class FileWrapper(object):  # pylint: disable=too-few-public-methods
    def __init__(self, workspace, file_definition):
        self.file_name = file_definition
        self.parser = None

        self._workspace = workspace

        self.populate()

    def populate(self):
        try:
            _sp = self.file_name.split('#')
            self.file_name = _sp[1]

            identifier = _sp[0]
            failures_only = False
            failures_count = False
            try:
                failures_only = _sp[0].split('-')[1] == Constants.FAILURES
                failures_count = _sp[0].split('-')[1] == Constants.FAILURES_COUNT
                identifier = _sp[0].split('-')[0]
            except (IndexError, AttributeError):
                pass

            self.parser = get_parser(identifier, self._workspace, failures_only, failures_count)
        except Exception:  # pylint: disable=broad-except
            pass



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
# pylint: disable=broad-except,protected-access
import os
import logging
import tempfile

import jsonpickle

from rapid.lib.in_memory_store import InMemoryStore

logger = logging.getLogger("rapid")

try:
    import uwsgi
except ImportError:
    uwsgi = InMemoryStore()


class StoreService(object):

    @staticmethod
    def get_executors():
        executors = []
        try:
            tmp_dir = tempfile.gettempdir()
            for filename in os.listdir(tmp_dir):
                try:
                    _sp = filename.split('-')
                    if _sp[0] == 'rapid':
                        try:
                            os.kill(int(_sp[2]), 0)
                            executors.append({'action_instance_id': _sp[1], 'pid': _sp[2]})
                        except OSError:
                            os.remove(os.path.join(tmp_dir, filename))
                except Exception:
                    pass
        except Exception as exception:
            logger.exception(exception)
        return executors

    @staticmethod
    def _get_tempfile_name(executor):
        return os.path.join(tempfile.gettempdir(), "rapid-{}-{}".format(executor.work_request.action_instance_id, executor.pid))

    @staticmethod
    def clear_executor(executor):
        try:
            os.remove(StoreService._get_tempfile_name(executor))
        except OSError:
            pass

    @staticmethod
    def save_executor(executor):
        try:
            with open(StoreService._get_tempfile_name(executor), 'w') as file_out:
                file_out.write("{}".format(executor.pid))
        except IOError as exception:
            logger.exception(exception)

    @staticmethod
    def save_clients(clients, app):
        try:
            uwsgi.cache_update("_rapidci_clients", jsonpickle.dumps(clients))
        except (AttributeError, NameError) as exception:
            logger.exception(exception)
            app.rapid_config.clients = clients

    @staticmethod
    def get_clients(app):
        """
        :param app:
        :type app:
        :return:
        :rtype: dict
        """
        try:
            return jsonpickle.loads(StoreService.get_key('_rapidci_clients'))
        except Exception:
            if not hasattr(app.rapid_config, 'clients'):
                app.rapid_config.clients = {}
            return app.rapid_config.clients

    @staticmethod
    def save_master_key(app, api_key):  # pylint: disable=unused-argument
        return StoreService.__set_key('_rapidci_master_key', jsonpickle.dumps(api_key))

    @staticmethod
    def get_master_key(app):
        try:
            return jsonpickle.loads(uwsgi.cache_get("_rapidci_master_key"))
        except Exception:
            return app.rapid_config._rapidci_master_key if hasattr(app.rapid_config, "_rapidci_master_key") else None

    @staticmethod
    def set_updating(app, updating=True):
        try:
            uwsgi.cache_update("_rapidci_updating", jsonpickle.dumps(updating))
        except Exception:
            app.rapid_config._rapidci_updating = updating

    @staticmethod
    def is_updating(app):
        try:
            return jsonpickle.loads(StoreService.get_key('_rapidci_updating'))
        except Exception:
            return app.rapid_config._rapidci_updating if hasattr(app.rapid_config, "_rapidci_updating") else False

    @staticmethod
    def check_for_pidfile(action_instance_id):
        try:
            pid_name = 'rapid-{}'.format(action_instance_id)
            for filename in os.listdir(tempfile.gettempdir()):
                if pid_name in filename:
                    return filename
        except OSError:
            pass
        return None

    @staticmethod
    def is_completing(action_instance_id):
        return StoreService.__is_by_key("_completing_{}".format(action_instance_id), 'true')

    @staticmethod
    def set_completing(action_instance_id):
        return StoreService.__set_key('_completing_{}'.format(action_instance_id), 'true')

    @staticmethod
    def clear_completing(action_instance_id):
        return StoreService.__clear_key("_completing_{}".format(action_instance_id))

    @staticmethod
    def set_calculating_workflow(pipeline_instance_id):
        return StoreService.__set_key('_calculating_{}'.format(pipeline_instance_id), "true")
        
    @staticmethod
    def is_calculating_workflow(pipeline_instance_id):
        return StoreService.__is_by_key('_calculating_{}'.format(pipeline_instance_id), 'true')

    @staticmethod
    def clear_calculating_workflow(pipeline_instance_id):
        return StoreService.__clear_key('_calculating_{}'.format(pipeline_instance_id))

    @staticmethod
    def __is_by_key(key, value):
        try:
            return value == StoreService.get_key(key)
        except Exception:
            pass
        return False

    @staticmethod
    def __set_key(key, value):
        try:
            uwsgi.cache_update(key, value)
            return True
        except Exception:
            pass
        return False

    @staticmethod
    def __clear_key(key):
        try:
            uwsgi.cache_del(key)
            return True
        except Exception:
            logger.info("FAILED TO clear cache-key: {}".format(key))
        return False

    @staticmethod
    def get_key(key):
        try:
            return uwsgi.cache_get(key)
        except Exception:
            return None



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

from rapid.lib import api_key_required, json_response
from rapid.lib.constants import Constants
from rapid.lib.framework.injectable import Injectable
from rapid.release.data.release_dal import ReleaseDal


class ReleaseController(Injectable):
    __injectables__ = {'release_dal': ReleaseDal}

    def __init__(self, release_dal):
        """
        :type release_dal: :class:`rapid.release.data.release_dal.ReleaseDal`
        :return:
        """
        self.release_dal = release_dal

    def set_status_step(self, release_id, step_id, status):
        return self.release_dal.set_release_step(release_id, step_id, status)

    def mark_step_for_commit(self, commit_identifier, step_custom_id, status_name):
        return self.release_dal.mark_step_for_commit(commit_identifier, step_custom_id, status_name)


class ReleaseRouter(Injectable):
    __injectables__ = {'release_controller': ReleaseController}

    def __init__(self, release_controller, flask_app):
        self.flask_app = flask_app
        self.release_controller = release_controller

    def configure_urls(self):
        if self.flask_app:
            self.flask_app.add_url_rule(Constants.get_api_url("releases/<int:release_id>/step/<path:step_id>/<path:status>"), "status_step", api_key_required(self.status_step), methods=["POST"])
            self.flask_app.add_url_rule(Constants.get_api_url("releases/<path:commit_identifier>/step/<path:step_custom_id>/<path:status_name>"),
                                        "mark_step_for_commit",
                                        api_key_required(self.mark_step_for_commit),
                                        methods=['POST'])

    @json_response(message="Something did not work.")
    def status_step(self, release_id, step_id, status):
        return self.release_controller.set_status_step(release_id, step_id, status)

    @json_response()
    def mark_step_for_commit(self, commit_identifier, step_custom_id, status_name):
        return self.release_controller.mark_step_for_commit(commit_identifier, step_custom_id, status_name)



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

from flask import Response

from rapid.lib.version import Version
from rapid.lib import api_key_required
from rapid.lib.utils import UpgradeUtil


class UpgradeController(object):

    def __init__(self, flask_app):
        self.flask_app = flask_app

    def configure_routing(self):
        self.flask_app.add_url_rule('/api/upgrade/<path:version>', 'upgrade_master', api_key_required(self.upgrade_master), methods=['POST'])

    def upgrade_master(self, version):
        worked = UpgradeUtil.upgrade_version(version, self.flask_app.rapid_config)
        return Response("It worked!" if worked else "It didn't work, version {} restored!".format(Version.get_version()), status=200 if worked else 505)



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

import logging
import threading
import time

from flask import Flask, Response

from rapid.master.data import run_db_downgrade, configure_db
from rapid.lib import setup_config_from_file, is_primary_worker, setup_status_route
from rapid.lib.framework.ioc import IOC
from .controllers import register_controllers
from .data import configure_data_layer, run_db_upgrades, create_revision

app = Flask("rapidci_master")
app.rapid_config = {'_is': 'master'}
handler = logging.StreamHandler()
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
handler.setFormatter(formatter)
# app.logger.addHandler(handler)  # pylint: disable=no-member
# app.logger.setLevel(logging.INFO)  # pylint: disable=no-member

logger = logging.getLogger("rapid")
logger.addHandler(handler)  # pylint: disable=no-member
logger.setLevel(logging.INFO)  # pylint: disable=no-member


# pylint: disable=broad-except

# Debug Code for sniffing out nplus one code!
# try:
#     # app.config['NPLUSONE_LOGGER'] = logging.getLogger('app.nplusone')
#     app.config['NPLUSONE_LOG_LEVEL'] = logging.ERROR
#     app.config['NPLUSONE_RAISE'] = True
#     from nplusone.ext.flask_sqlalchemy import NPlusOne
#     NPlusOne(app)
# except Exception:
#     pass


@app.errorhandler(500)
def internal_error(exception):
    response = Response(status=500, content_type='application/json')
    response.data = _to_dict(exception)
    app.logger.error(exception.message)  # pylint: disable=no-member
    return response


def _to_dict(exception):
    _rv = dict()
    _rv['message'] = exception.message if hasattr(exception, 'message') else 'An exception has occurred'
    return _rv


def configure_application(flask_app, args, manual_db_upgrade=False):
    IOC.register_global('flask_app', flask_app)
    setup_config_from_file(flask_app, args)
    configure_db()
    configure_sub_modules(flask_app, args)
    configure_data_layer(flask_app)
    register_controllers(flask_app)
    setup_status_route(flask_app)

    if args.static_file_dir:
        flask_app.rapid_config.static_file_directory = args.static_file_dir

    if args.basic_auth:
        flask_app.rapid_config.set_basic_auth(args.basic_auth)

    if is_primary_worker() and not manual_db_upgrade:
        if args.db_downgrade:
            run_db_downgrade(flask_app, args.db_downgrade)
        else:
            run_db_upgrades(flask_app)
            setup_queue_thread(flask_app)
    load_extensions(flask_app)


def load_extensions(flask_app):
    from rapid.lib.framework.ioc import IOC
    from rapid.extensions.extension_loader import ExtensionLoader
    extension_loader = IOC.get_class_instance(ExtensionLoader)
    extension_loader.load_extensions(flask_app)


def create_migration_script(flask_app, migration):
    create_revision(flask_app, migration)


def configure_sub_modules(flask_app, args):  # pylint: disable=unused-argument
    logger.info("------- Searching for and starting modules --------")
    modules = {}
    logger.info("-- Registering IOC Globals")
    for name in ['ci', 'qa', 'release', 'workflow']:
        try:
            module = __import__("rapid.{}".format(name), fromlist=['configure_module', 'register_ioc_globals'])
            modules[name] = module
            module.register_ioc_globals(flask_app)
            logger.info("    Registered: {}".format(name))
        except Exception:
            import traceback
            traceback.print_exc()

    logger.info("-- Configuring modules")
    for name, item in list(modules.items()):  # pylint: disable=unused-variable
        try:
            modules[name].configure_module(flask_app)
            logger.info("    Configured: {}".format(name))
        except Exception:
            import traceback
            traceback.print_exc()
            logger.warning("NOT LOADED! Module {} did not load.".format(name))

    logger.info("------- Done --------")


def setup_queue_thread(flask_app):
    if flask_app.rapid_config.queue_manager:
        import os
        logger.info("Setting up master queue thread: [{}]".format(os.getpid()))
        thread = threading.Thread(target=run_queue, args=[flask_app])
        thread.daemon = True
        thread.start()


def run_queue(flask_app):
    from rapid.lib.store_service import StoreService
    from rapid.workflow.queue import Queue
    with flask_app.app_context():
        queue = IOC.get_class_instance(Queue)  # type: Queue
        while True:
            clients = []
            try:
                clients = StoreService.get_clients(flask_app)
            except:
                pass

            try:
                queue.process_queue(clients)
            except Exception as exception:
                logger.error(exception)

            try:
                queue.verify_still_working(clients)
            except Exception as exception:
                logger.error(exception)

            try:
                queue.reconcile_pipeline_instances()
            except Exception as exception:
                logger.error(exception)

            try:
                queue.reconcile_releases()
            except Exception as exception:
                logger.exception(exception)

            filtered_clients = {name: client for name, client in list(clients.items()) if not hasattr(client, 'no-longer-active')}

            if filtered_clients != clients:
                StoreService.save_clients(filtered_clients, flask_app)
            time.sleep(flask_app.rapid_config.queue_time)



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

import glob
import logging
import os
import platform
import shutil
import subprocess
import tempfile
import threading

from rapid import testmapper
from rapid.client.communicator.client_communicator import ClientCommunicator
from rapid.lib.communication import Communication
from rapid.lib.constants import Constants
from rapid.lib.exceptions import ThresholdException, ResultsFileNotFoundException, ResultsFileNotParsedException
from rapid.lib.features import Features
from rapid.lib.store_service import StoreService
from rapid.lib.utils import deep_merge, OSUtil

rapid_logger = logging.getLogger('rapid')

# pylint: disable=broad-except, too-many-instance-attributes


class Executor(object):
    def __init__(self, work_request, master_uri, logger=None, workspace=None, quarantine=None, verify_certs=True, rapid_config=None):
        """

        :param work_request:
        :param master_uri:
        :param logger:
        :param workspace:
        :param quarantine:
        :param verify_certs:
        :type rapid_config: :class:
        :return:
        """
        self.thread = None
        self.reading_thread = None
        self.thread_id = None
        self.read_pid = None
        self.work_request = work_request
        self.child_process = None
        self.workspace = workspace if workspace is not None else os.path.join(tempfile.gettempdir(), 'rapid', 'workspace')
        self.master_uri = master_uri
        self.reading_thread = None
        self.logger = logger
        self.status_overrides = {}
        self.results_files = []
        self.parameter_files = []
        self.stats_files = []
        self.quarantine = quarantine
        self.failure_threshold = 0
        self.verify_certs = verify_certs
        self.rapid_config = rapid_config
        self.analyze_tests = None
        self.pid = None
        self._file_handler = None

        if self.logger is None:
            self.logger = self._create_logger()

    def __getstate__(self):
        state = self.__dict__.copy()
        for attribute in ['logger', 'thread', 'reading_thread', 'child_process', 'rapid_config', '_file_handler']:
            if attribute in state:
                del state[attribute]

        return state

    def verify_work_request(self):
        if not self.work_request.action_instance_id:
            raise Exception("Invalid action_instance_id")
        if not self.work_request.executable:
            raise Exception("Executable not set, nothing to run.")
        if not self.work_request.pipeline_instance_id:
            raise Exception("No pipline_instance_id set.")
        if not self.work_request.cmd:
            raise Exception("No command was set.")

    def _create_logger(self):
        log_file_name = self._get_log_file_name()
        if not log_file_name:
            return rapid_logger

        logger = logging.getLogger("rapid-executor")
        logger.setLevel(logging.INFO)

        self._file_handler = logging.FileHandler(log_file_name)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        self._file_handler.setFormatter(formatter)
        logger.addHandler(self._file_handler)
        
        return logger

    def _get_log_file_name(self):
        if self.rapid_config and self.rapid_config.log_to_directory:
            try:
                os.makedirs(os.path.join(self.rapid_config.log_to_directory), exist_ok=True)
            except:  # pylint: disable=bare-except
                pass
            return os.path.join(self.rapid_config.log_to_directory, '{}.log'.format(self.work_request.action_instance_id))
        return None

    def start(self, threaded=True):
        """
        Spin off a process that will execute and run the given work request.
        :return:
        """
        if self._get_log_file_name():
            rapid_logger.info("Running: {}".format(self._get_log_file_name()))

        if threaded:
            self.thread = threading.Thread(target=self._start_child_process)
            self.thread.daemon = True
            self.thread.start()
        else:
            self._start_child_process()

    def _start_child_process(self):
        """
        Start a child process with the command.
        :return:
        """
        communicator = ClientCommunicator(self.master_uri,
                                          self.quarantine,
                                          verify_certs=self.verify_certs,
                                          get_files_auth=self.rapid_config.get_files_basic_auth)

        # Cleanup first
        self.workspace = OSUtil.path_join(self.workspace, str(self.work_request.action_instance_id))
        self.clean_workspace()

        env = self.get_environment()
        try:
            command = self.get_command(communicator)
            command.extend(self.get_arguments())

            # Check and download required files that will execute in command.
            try:
                self._download_required_files([self.work_request.executable], communicator, self.logger)
            except Exception as exception:
                self.logger.exception(exception)
                raise exception

            stdout = None
            stderr = None
            if self.logger:
                stdout = subprocess.PIPE
                stderr = subprocess.STDOUT

            self.child_process = subprocess.Popen(command, cwd=self.workspace, env=env, stderr=stderr, stdout=stdout)
            self.pid = self.child_process.pid

            StoreService.save_executor(self)

            # wait for the process to finish
            self.reading_thread = threading.Thread(target=self._read_process_output, args=(self.logger, self.child_process))
            self.reading_thread.daemon = True
            self.reading_thread.start()

            self.child_process.wait()

            self._finalize_run(communicator)
        except (BaseException, Exception) as exception:
            self.logger.exception(exception)
            communicator.send_done(self.work_request.action_instance_id, 'FAILED', None, None, None, self.logger)
        finally:
            self.clean_workspace()

    def _finalize_run(self, communicator):
        # Send results to master
        # Parse and Send parameters
        self.check_for_dynamic_config_file()

        status = self._get_status(self.child_process.returncode)
        results = None
        try:
            results = self._get_results(self.workspace, self.results_files)
            if results:
                if Constants.RESULTS_SUMMARY in results and 'FAILED' in results[Constants.RESULTS_SUMMARY] and results[Constants.RESULTS_SUMMARY]['FAILED'] > self.failure_threshold:
                    raise ThresholdException("Failure threshold was exceeded")

        except ThresholdException as exception:
            status = 'UNSTABLE'
        except ResultsFileNotFoundException as exception:
            self.logger.error("Expected ResultsFile was not where specified.")
            status = 'FAILED'
        except ResultsFileNotParsedException:
            self.logger.error("Expected ResultsFile was not parsed.")
            status = 'FAILED'
        finally:
            parameters = Executor._get_parameters(self.workspace, self.parameter_files)
            stats = Executor._get_stats(self.workspace, self.stats_files)

            self.logger.info("Sending the Done command!")
            communicator.send_done(self.work_request.action_instance_id, status, parameters, stats, results, self.logger)

        try:
            name_map = self._get_name_map(results)
            test_analysis = self._get_test_analysis(self.workspace, self.analyze_tests, name_map)

            if test_analysis is not None:
                communicator.send_test_analysis(self.work_request.pipeline_instance_id, test_analysis, self.logger)
        except Exception as exception:
            self.logger.error(exception)

        if self._file_handler:
            self._file_handler.close()
            self.logger.removeHandler(self._file_handler)
        # Clear only after having sent the done.
        StoreService.clear_executor(self)

        if self._get_log_file_name():
            rapid_logger.info("Finished: {}".format(self._get_log_file_name()))

    def _get_name_map(self, results):
        name_map = {}
        if results:
            for name in list(results.keys()):
                _sp = name.split('~')
                last = _sp[-1]
                name_map[last] = name
        return name_map

    @staticmethod
    def _get_parameters(workspace, parameter_files):
        if parameter_files:
            parameters = {}
            for glob_tmp in parameter_files:
                for parameter_file in glob.glob(os.path.join(workspace, glob_tmp)):
                    with open(parameter_file) as glob_file:
                        lines = glob_file.readlines()
                        parameters.update(Executor._convert_to_dict(lines))

            return parameters
        return None

    @staticmethod
    def _get_stats(workspace, stats_files):
        if stats_files:
            stats = {}
            for glob_tmp in stats_files:
                try:
                    for stats_file in glob.glob(os.path.join(workspace, glob_tmp)):
                        with open(stats_file) as glob_file:
                            lines = glob_file.readlines()
                            stats.update(Executor._convert_to_dict(lines))
                except Exception:  # No need to worry about failing stats files.
                    pass
            return stats
        return None

    @staticmethod
    def _convert_to_dict(lines):
        return dict((k.strip(), v.strip()) for k, v in
                    (item.split('=', 1) for item in lines))

    def _get_results(self, workspace, results_files):  # pylint: disable=too-many-locals
        if results_files:
            # Wire in the registered parsers.
            results = {}
            from .parsers import parse_file, FileWrapper

            for file_glob in results_files:
                wrapper = FileWrapper(self.workspace, file_glob)

                selected_parser = wrapper.parser

                file_was_read = False
                for results_file in glob.glob("{}/{}".format(workspace, wrapper.file_name)):
                    file_was_read = True
                    with open(results_file) as glob_file:
                        try:
                            lines = glob_file.read().splitlines(True)
                            parsed_results = parse_file(lines, selected_parser)

                            summary = parsed_results[Constants.RESULTS_SUMMARY] if Constants.RESULTS_SUMMARY in parsed_results else None
                            if summary is not None:
                                if Constants.RESULTS_SUMMARY not in results:
                                    results[Constants.RESULTS_SUMMARY] = summary
                                else:
                                    for (key, value) in summary.items():
                                        results[Constants.RESULTS_SUMMARY][key] += value
                                del parsed_results[Constants.RESULTS_SUMMARY]

                            results.update(parsed_results)
                        except Exception:
                            raise ResultsFileNotParsedException("Result file did not parse.")

                if not file_was_read:
                    raise ResultsFileNotFoundException("Results file not read.")

            return results
        return None

    def _get_status(self, return_code):
        status = 'SUCCESS'
        if return_code > 0:
            status = 'FAILED'
            status = self.status_overrides[return_code] if return_code in self.status_overrides else status
        elif return_code < 0:
            status = 'CANCELED'
        return status

    def _read_process_output(self, logger, child_process):
        try:
            for line in iter(child_process.stdout.readline, b''):
                self._log(self.work_request.action_instance_id, line.strip(), logger)
            self._log(self.work_request.action_instance_id, "Cleaning up thread.", logger)
        except Exception as exception:
            self._log(self.work_request.action_instance_id, str(exception), logger)

    @staticmethod
    def _get_test_analysis(workspace, test_analysis, name_map):
        results = {}
        if isinstance(test_analysis, str):
            for dirname in glob.glob("{}/{}".format(workspace, test_analysis)):
                results = deep_merge(results, testmapper.process_directory(workspace, dirname, name_map=name_map))
        elif isinstance(test_analysis, list):
            for glob_tmp in test_analysis:
                for filename in glob.glob("{}/{}".format(workspace, glob_tmp)):
                    results = deep_merge(results, testmapper.process_directory(workspace, filename, name_map=name_map))
        return results

    @staticmethod
    def _log(action_instance_id, message, logger):
        try:
            if isinstance(message, bytes):
                message = message.decode('utf-8')
            try:
                logger.info(u"__RCI_{}__ - {} - {}".format(action_instance_id, os.getpid(), message))
            except:
                logger.info("__RCI_{}__ - {} - {}".format(action_instance_id, os.getpid(), message))  # Python 3 no unicode cast needed.
        except AttributeError:
            #  Thread race condition on windows os will occasionally be None
            pass

    def get_environment(self):
        env = {}
        for key in os.environ:
            env[key] = os.environ[key]

        if self.work_request.environment:
            for key, value in self.work_request.environment.items():
                try:
                    if type(key) != str:
                        key = key.decode('utf-8')
                    if type(value) != str:
                        value = value.decode('utf-8')

                    env[key] = value
                except (AttributeError, UnicodeEncodeError):
                    print("Failed to encode: {} => {}".format(key, value))

        env['PYTHONUNBUFFERED'] = "true"
        env['pipeline_instance_id'] = str(self.work_request.pipeline_instance_id)
        env['action_instance_id'] = str(self.work_request.action_instance_id)
        env['workflow_instance_id'] = str(self.work_request.workflow_instance_id)
        env['slice'] = str(self.work_request.slice)
        env['RAPID_FEATURES'] = ",".join(Features.get_enabled_features())
        env['WORKSPACE'] = self.workspace

        return env

    def get_command(self, communicator):
        """
        :type communicator: :class:`rapid.client.communicator.client_communicator.ClientCommunicator`
        :return:
        """
        command = []
        if self.work_request.cmd:
            command.append(self.work_request.cmd)
        if self.work_request.executable:
            command.append(communicator.get_downloaded_file_name(self.workspace,
                                                                 self._get_executable_name(self.work_request.executable)))
        return command

    def get_arguments(self):
        arguments = []
        if self.work_request.args:
            arguments.extend(self.work_request.args.split(' '))
        return arguments

    def check_for_dynamic_config_file(self, filename=None):
        config_path = os.path.join(self.workspace, 'rapid_config') if filename is None else filename
        if os.path.isfile(config_path):
            try:
                with(open(config_path, 'r')) as tmp_file:
                    self.verify_file_lines(tmp_file.readlines(), self.logger)
            except Exception:
                pass

    def verify_file_lines(self, lines, logger):
        for line in lines:
            self.verify_lines(line, logger)

    def _normalize_workspace(self):
        replacement = ''
        if os.sep == '/':
            replacement = '\\'
        else:
            replacement = '/'
        return self.workspace.replace(replacement, os.sep)

    def clean_workspace(self):
        normalized_workspace = self._normalize_workspace()
        if os.path.isdir(normalized_workspace):
            try:
                Executor._log(self.work_request.action_instance_id, "{} - removing workspace".format(normalized_workspace), self.logger)
                if platform.system() == 'Windows':
                    os.system('rmdir /S /Q {}'.format(normalized_workspace))
                else:
                    shutil.rmtree(normalized_workspace, ignore_errors=True)
            except Exception:
                pass
        else:
            try:
                os.makedirs(self.workspace)
            except Exception as exception:
                Executor._log(self.work_request.action_instance_id, "{} - Workspace was unable to be created.".format(self.workspace), self.logger)
                self.logger.exception(exception)

    @staticmethod
    def _get_executable_name(file_name):
        if file_name and file_name.startswith(Communication.REMOTE_FILE):
            return file_name.split(Communication.REMOTE_FILE, 2)[1]
        return file_name

    def _download_required_files(self, file_names, communicator, logger, is_from_file=False, files_downloaded=None):
        """
        Download the file, if a remote file, and then determine if there are any required files that need to be
        downloaded.
        :param list file_names: list of file names
        :param ClientCommunicator communicator: Communicator to download
        :return:
        """
        if not isinstance(file_names, list):
            raise Exception("file_names is not a list")

        files_downloaded = [] if files_downloaded is None else files_downloaded
        if file_names:
            for file_name in file_names:
                if Communication.REMOTE_FILE in file_name or is_from_file:
                    Executor._log(self.work_request.action_instance_id, "Downloading file: {}".format(file_name),
                                  logger)
                    true_file_name = Executor._get_executable_name(file_name)
                    files_downloaded.append(file_name)
                    with open(communicator.get_file(self.workspace, true_file_name, logger), 'r') as tmp_file:
                        lines = tmp_file.readlines()
                        self._download_by_lines(lines,
                                                communicator,
                                                logger,
                                                files_downloaded)

    def verify_lines(self, line, logger):
        try:
            results_files = Executor._get_results_files(line)
            self.results_files.extend(results_files)
        except Exception as exception:
            logger.debug(exception)
        try:
            parameters_file = Executor._get_parameters_files(line)
            self.parameter_files.extend(parameters_file)
        except Exception as exception:
            logger.debug(exception)
        try:
            stats_file = Executor._get_stats_files(line)
            self.stats_files.extend(stats_file)
        except Exception as exception:
            logger.debug(exception)

        try:
            threshold = Executor._get_failure_threshold(line)[0]
            self.failure_threshold = int(threshold)
        except Exception as exception:
            logger.debug(exception)

        try:
            self.status_overrides.update(Executor._get_status_overrides(line))
        except Exception as exception:
            logger.debug(exception)

        try:
            if self.analyze_tests is None:
                self.analyze_tests = Executor._get_analyze_tests(line)
            else:
                self.analyze_tests.extend(Executor._get_analyze_tests(line))
        except Exception as exception:
            logger.debug(exception)

    def _download_by_lines(self, lines, communicator, logger, files_downloaded=None):
        """
        Download files by recursing list of lines.
        :param list lines:
        :param ClientCommunicator communicator:
        :return:
        """
        files_downloaded = [] if files_downloaded is None else files_downloaded
        for line in lines:
            try:
                self._download_required_files(Executor._get_remote_files(line),
                                              communicator,
                                              logger,
                                              is_from_file=True,
                                              files_downloaded=files_downloaded)
            except Exception:
                logger.debug("Line returned nothing for required files: {}".format(line))

            self.verify_lines(line, logger)

    @staticmethod
    def _get_analyze_tests(line):
        """
        Splits a line for Constants.REQUIRED_FILES and ',' and returns the split.
        :param line: line from a file to split
        :return: list from the split
        :raises  IndexError if line doesn't contain Constants.REQUIRED_FILES
        """
        return Executor._get_split_string(line, Constants.ANALYZE_TESTS)


    @staticmethod
    def _get_remote_files(line):
        """
        Splits a line for Constants.REQUIRED_FILES and ',' and returns the split.
        :param line: line from a file to split
        :return: list from the split
        :raises  IndexError if line doesn't contain Constants.REQUIRED_FILES
        """
        return Executor._get_split_string(line, Constants.REQUIRED_FILES)

    @staticmethod
    def _get_results_files(line):
        """
        Splits a line from Constants.RESULTS_FILES and ',' and returns the split.
        :param line: line from a file to split
        :return: list from the split
        :raises IndexError if line doesn't contain Constants.RESULTS
        """
        return Executor._get_split_string(line, Constants.RESULTS)

    @staticmethod
    def _get_stats_files(line):
        """
        Splits a line from Constants.STATS_FILES and ',' and returns the split
        :param line: line from a file to split
        :return: list from a the split
        """
        return Executor._get_split_string(line, Constants.STATS)

    @staticmethod
    def _get_parameters_files(line):
        """
        Splits a line from Constants.PARAMETERS and ',' and returns the split
        :param line: line to split
        :return: list from the split
        :raises IndexError if line doesn't contain Constants.PARAMETERS
        """
        return Executor._get_split_string(line, Constants.PARAMETERS)

    @staticmethod
    def _get_failure_threshold(line):
        return Executor._get_split_string(line, Constants.THRESHOLD)

    @staticmethod
    def _get_status_overrides(line):
        overrides = {}
        try:
            for override in Executor._get_split_string(line, Constants.STATUS_OVERRIDE):
                _sp = override.split(':')
                if len(_sp) > 1:
                    overrides[int(_sp[0])] = _sp[1]
        except Exception:
            pass
        return overrides

    @staticmethod
    def _get_split_string(line, constant):
        return line.strip().split(constant)[1].split(",")




from abc import ABCMeta, abstractmethod


class QueueHandler(object):
    _GRAIN_SPLIT = '://'
    __metaclass__ = ABCMeta

    def __init__(self, rapid_config):
        """
        Parameters
        ----------
        rapid_config: rapid.master.master_configuration.MasterConfiguration
        """
        self.rapid_config = rapid_config

    @abstractmethod
    def process_work_request(self, work_request, clients):
        yield

    @abstractmethod
    def can_process_work_request(self, work_request):
        # type: (WorkRequest) -> bool
        yield

    @abstractmethod
    def process_action_instance(self, action_instance, clients):
        # type: (dict, list) -> bool
        yield

    @abstractmethod
    def can_process_action_instance(self, action_instance):
        # type: (dict) -> bool
        yield

    @abstractmethod
    def cancel_worker(self, action_instance):
        # type: (dict) -> bool
        yield

    def _get_grain_type_split(self, grain):
        return grain.split(self._GRAIN_SPLIT, 1)



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
# pylint: disable=broad-except
import os
import re
import sys
import hmac
import logging
import datetime

from hashlib import sha1

from rapid.lib.version import Version

logger = logging.getLogger("rapid")


def deep_merge(_a, _b, path=None):
    "merges b into a"
    if path is None:
        path = []
    for key in _b:
        if key in _a:
            if isinstance(_a[key], dict) and isinstance(_b[key], dict):
                deep_merge(_a[key], _b[key], path + [str(key)])
            elif isinstance(_a.get(key, None), list) and isinstance(_b.get(key, None), list):
                _a[key].extend(_b[key])
            elif _a[key] == _b[key]:
                pass  # same leaf value
            else:
                raise Exception('Conflict at %s' % '.'.join(path + [str(key)]))
        else:
            if isinstance(_a.get(key, None), list) and isinstance(_b.get(key, None), list):
                _a[key].extend(_b[key])
            else:
                _a[key] = _b[key]
    return _a


class RoutingUtil(object):

    @staticmethod
    def is_valid_request(sent_api_key, api_key):
        return str(sent_api_key) == str(api_key)

    @staticmethod
    def get_cihub_signature(secret_key, data):
        return hmac.new(secret_key.encode('ascii', 'ignore'), data, sha1).hexdigest()


class ORMUtil(object):
    @staticmethod
    def get_filtered_query(query, _filter, clazz):  # pylint: disable=too-many-locals
        if _filter is not None:
            joined_classes = []
            for token in _filter.split(':=:'):
                _m = re.match('(.*)(_eq_|_ne_|_gt_|_lt_|_in_|_ge_|_le_|_like_)(.*)', token)
                if not _m:
                    return query

                group1 = _m.group(1).split('.')
                _op = _m.group(2).replace('_', '')
                value = _m.group(3)

                if len(group1) > 1:
                    final_class = None
                    final_column = None
                    current_class = clazz
                    while len(group1) > 1:

                        nested_attribute = group1[1]
                        column_name = group1.pop(0)

                        attr = getattr(current_class, column_name, None)
                        final_class = attr.property.mapper.class_
                        final_column = getattr(final_class, nested_attribute, None)

                        parent_name = final_column.parent.mapper.class_
                        if parent_name not in joined_classes:
                            query = query.join(final_column.parent.mapper.class_)
                            joined_classes.append(parent_name)

                        current_class = final_column.parent.mapper.class_

                    if final_class is not None and final_column is not None:
                        query = ORMUtil.get_embedded_filter(final_class, final_column.name, _op, value, query)
                else:
                    column_name = group1[0]
                    if hasattr(clazz, column_name):
                        query = ORMUtil.get_column_filter(clazz, column_name, _op, value, query)
                    else:
                        logger.info("Not found: {}".format(_m.group(1)))
        return query

    @staticmethod
    def get_column_filter(clazz, column_name, _op, value, query):
        if hasattr(clazz, column_name):
            query = ORMUtil.get_embedded_filter(clazz, column_name, _op, value, query)
            return query
        raise Exception("Missing attribute")

    @staticmethod
    def get_embedded_filter(clazz, column_name, _op, value, query):
        if hasattr(clazz, column_name):
            column = getattr(clazz, column_name, None)
            filt = None
            if _op == 'in':
                filt = column.in_(value.split(','))
            else:
                attr = None
                try:
                    for _e in ['%s', '%s_', '__%s__']:
                        if hasattr(column, _e % _op):
                            attr = _e % _op
                except Exception as exception:
                    logger.info("\n\n")
                    logger.error(exception)

                if value == 'null':
                    value = None

                if column.type.python_type == bool:
                    value = value == 'True'
                elif column.type.python_type == datetime.datetime:
                    try:
                        value = datetime.datetime.utcfromtimestamp(float(value))  # Python timestamp
                    except ValueError:
                        value = datetime.datetime.utcfromtimestamp(float(value / 1e3))  # Traditional linux timestamp
                filt = getattr(column, attr)(value)
            query = query.filter(filt)
        return query


class UpgradeUtil(object):

    @staticmethod
    def upgrade_version(version, configuration, attempted_reinstall=False):
        if UpgradeUtil._install(version, configuration):
            try:
                import uwsgi
                uwsgi.reload()
            except ImportError:
                import traceback
                traceback.print_exc()
                return False
            return True
        if not attempted_reinstall:
            if UpgradeUtil.upgrade_version(Version.get_version(), configuration, True):
                return False

        raise Exception("Server was unable to upgrade nor restore to previous version!")

    @staticmethod
    def _install(version, configuration):
        version = version.split(';')[0]
        assert re.search(r'[^a-zA-Z\-0-9.]', version) is None, "Invalid Version string."
        
        try:
            logger.info("installing version: {}".format(version))
            return_code = os.system("{}/bin/pip install -i {} {} rapid-framework=={}".format(sys.prefix, configuration.install_uri, configuration.install_options, version))
            return return_code == 0
        except Exception as exception:
            logger.error(exception)
            return False


class OSUtil:

    @staticmethod
    def separator():
        return os.getenv('os_path_override', os.path.sep)

    @staticmethod
    def path_join(*args):
        """
        :param args: List[str]
        :return:
        """
        return OSUtil.separator().join(args)



"""
Copyright (c) 2015 Michael Bright and Bamboo HR LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

qa_products

Revision ID: e66b5abaa8c2
Revises: 593023a41b6d
Create Date: 2016-10-18 02:27:03.604303

"""

# revision identifiers, used by Alembic.
from rapid.qa.data.models import QaTestType

revision = 'e66b5abaa8c2'
down_revision = '593023a41b6d'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table('qa_test_types',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=100), nullable=False),
    sa.Column('active', sa.Boolean(), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_qa_test_types_id'), 'qa_test_types', ['id'], unique=False)

    op.create_table('qa_products',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=250), nullable=False),
    sa.Column('vcs_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['vcs_id'], ['vcs.id'], ),
    sa.PrimaryKeyConstraint('id'),
    mysql_engine='InnoDB'
    )
    op.create_index(op.f('ix_qa_products_id'), 'qa_products', ['id'], unique=False)
    op.create_index(op.f('ix_qa_products_name'), 'qa_products', ['name'], unique=False)
    op.create_index(op.f('ix_qa_products_vcs_id'), 'qa_products', ['vcs_id'], unique=False)

    op.add_column('qa_areas', sa.Column('product_id', sa.Integer(), nullable=True))
    op.create_index(op.f('ix_qa_areas_product_id'), 'qa_areas', ['product_id'], unique=False)

    if 'sqlite' != op.get_context().dialect.name:
        op.create_foreign_key(None, 'qa_areas', 'qa_products', ['product_id'], ['id'])

    op.add_column('qa_behavior_points', sa.Column('product_id', sa.Integer(), nullable=True))
    op.create_index(op.f('ix_qa_behavior_points_product_id'), 'qa_behavior_points', ['product_id'], unique=False)

    if 'sqlite' != op.get_context().dialect.name:
        op.create_foreign_key(None, 'qa_behavior_points', 'qa_products', ['product_id'], ['id'])

    op.add_column('qa_features', sa.Column('product_id', sa.Integer(), nullable=True))
    op.create_index(op.f('ix_qa_features_product_id'), 'qa_features', ['product_id'], unique=False)

    if 'sqlite' != op.get_context().dialect.name:
        op.create_foreign_key(None, 'qa_features', 'qa_products', ['product_id'], ['id'])

    if 'sqlite' != op.get_context().dialect.name:
        op.alter_column('qa_test_mappings', 'feature_id',
                   existing_type=sa.INTEGER(),
                   nullable=True)
    op.create_index(op.f('ix_qa_test_mappings_area_id'), 'qa_test_mappings', ['area_id'], unique=False)
    op.create_index(op.f('ix_qa_test_mappings_behavior_id'), 'qa_test_mappings', ['behavior_id'], unique=False)
    op.create_index(op.f('ix_qa_test_mappings_feature_id'), 'qa_test_mappings', ['feature_id'], unique=False)

    op.add_column('qa_tests', sa.Column('qa_test_type_id', sa.Integer(), nullable=True))
    op.create_index(op.f('ix_qa_tests_qa_test_type_id'), 'qa_tests', ['qa_test_type_id'], unique=False)

    if 'sqlite' != op.get_context().dialect.name:
        op.create_foreign_key(None, 'qa_tests', 'qa_test_types', ['qa_test_type_id'], ['id'])

    op.create_index(op.f('ix_stage_instances_created_date'), 'stage_instances', ['created_date'], unique=False)
    op.create_index(op.f('ix_stage_instances_end_date'), 'stage_instances', ['end_date'], unique=False)
    op.create_index(op.f('ix_stage_instances_start_date'), 'stage_instances', ['start_date'], unique=False)

    op.create_index(op.f('ix_workflow_instances_created_date'), 'workflow_instances', ['created_date'], unique=False)
    op.create_index(op.f('ix_workflow_instances_end_date'), 'workflow_instances', ['end_date'], unique=False)
    op.create_index(op.f('ix_workflow_instances_start_date'), 'workflow_instances', ['start_date'], unique=False)
    ### end Alembic commands ###

    op.bulk_insert(QaTestType.__table__, [
        {"id": 1, "name": "Unit", "active": True},
        {"id": 2, "name": "Integration", "active": True},
        {"id": 3, "name": "Selenium", "active": True},
        {"id": 4, "name": "Smoke", "active": True},
    ])


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f('ix_workflow_instances_start_date'), table_name='workflow_instances')
    op.drop_index(op.f('ix_workflow_instances_end_date'), table_name='workflow_instances')
    op.drop_index(op.f('ix_workflow_instances_created_date'), table_name='workflow_instances')

    op.drop_index(op.f('ix_stage_instances_start_date'), table_name='stage_instances')
    op.drop_index(op.f('ix_stage_instances_end_date'), table_name='stage_instances')
    op.drop_index(op.f('ix_stage_instances_created_date'), table_name='stage_instances')

    op.drop_constraint(None, 'qa_tests', type_='foreignkey')
    op.drop_index(op.f('ix_qa_tests_qa_test_type_id'), table_name='qa_tests')
    op.drop_column('qa_tests', 'qa_test_type_id')

    op.drop_index(op.f('ix_qa_test_mappings_feature_id'), table_name='qa_test_mappings')
    op.drop_index(op.f('ix_qa_test_mappings_behavior_id'), table_name='qa_test_mappings')
    op.drop_index(op.f('ix_qa_test_mappings_area_id'), table_name='qa_test_mappings')

    if 'sqlite' != op.get_context().dialect.name:
        op.alter_column('qa_test_mappings', 'feature_id',
                   existing_type=sa.INTEGER(),
                   nullable=False)

    op.drop_constraint(None, 'qa_features', type_='foreignkey')
    op.drop_index(op.f('ix_qa_features_product_id'), table_name='qa_features')
    op.drop_column('qa_features', 'product_id')

    op.drop_constraint(None, 'qa_behavior_points', type_='foreignkey')
    op.drop_index(op.f('ix_qa_behavior_points_product_id'), table_name='qa_behavior_points')
    op.drop_column('qa_behavior_points', 'product_id')

    op.drop_constraint(None, 'qa_areas', type_='foreignkey')
    op.drop_index(op.f('ix_qa_areas_product_id'), table_name='qa_areas')
    op.drop_column('qa_areas', 'product_id')

    op.drop_index(op.f('ix_qa_products_vcs_id'), table_name='qa_products')
    op.drop_index(op.f('ix_qa_products_name'), table_name='qa_products')
    op.drop_index(op.f('ix_qa_products_id'), table_name='qa_products')
    op.drop_table('qa_products')

    op.drop_index(op.f('ix_qa_test_types_id'), table_name='qa_test_types')
    op.drop_table('qa_test_types')
    ### end Alembic commands ###



"""
Copyright (c) 2015 Michael Bright and Bamboo HR LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from rapid.lib.wrapping_thread_pool import WrappingThreadPool


class ResultsSerializer(object):

    @staticmethod
    def serialize_results(results, allowed_children=None, previous_children=None, thread_count=5):
        with WrappingThreadPool(thread_count) as pool:
            return pool.map(ResultsSerializer._convert, ResultsSerializer._get_results_generator(results, allowed_children, previous_children))

    @staticmethod
    def _convert(_x):
        (result, allowed_children, previous_children) = _x
        return result.serialize(allowed_children, previous_children)

    @staticmethod
    def _get_results_generator(results, allowed_children, previous_children):
        for qa_test_mapping in results:
            yield (qa_test_mapping, allowed_children, previous_children)



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

from rapid.lib.framework.ioc import IOC

registered_dals = []


def get_dal(model):
    global registered_dals  # pylint: disable=global-statement
    for dal in registered_dals:
        if dal.is_serviceable(model):
            return dal
    return None


def setup_dals(flask_app):
    from rapid.master.data.database.dal.general_dal import GeneralDal
    from rapid.workflow.data.dal.pipeline_dal import PipelineDal

    for dal in [PipelineDal, GeneralDal]:
        tmp = IOC.get_class_instance(dal)
        tmp.register_url_rules(flask_app)
        registered_dals.append(tmp)



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

from rapid.lib.framework.injectable import Injectable
from rapid.workflow.action_dal import ActionDal
from rapid.workflow.data.dal.pipeline_dal import PipelineDal
from rapid.workflow.data.dal.report_dal import ReportDal


class ActionService(Injectable):
    __injectables__ = {'pipeline_dal': PipelineDal, 'action_dal': ActionDal, 'report_dal': ReportDal}

    def __init__(self, pipeline_dal, action_dal, report_dal):
        """

        :param pipeline_dal:
        :type pipeline_dal: PipelineDal
        :param action_dal:
        :type action_dal: ActionDal
        :param report_dal:
        :type report_dal: ReportDal
        """
        self.pipeline_dal = pipeline_dal
        self.action_dal = action_dal
        self.report_dal = report_dal

    def start_pipeline_via_repo(self, repo, json_data):
        """
        Start a pipeline via repository URL
        :rtype: dict
        """
        return self.pipeline_dal.start_pipeline_instance_via_reponame(repo, json_data)

    def get_status_by_name(self, name, session=None):
        return self.action_dal.get_status_by_name(name, session)

    def start_pipeline_by_id(self, pipeline_id, json_data):
        return self.pipeline_dal.start_pipeline_instances_via_pipeline_id(pipeline_id, json_data=json_data)

    def get_pipeline_instance_by_id(self, pipeline_instance_id):
        return self.pipeline_dal.get_pipeline_instance_by_id(pipeline_instance_id)

    def get_custom_report(self, report_name):
        return self.report_dal.get_canned_report(report_name)

    def list_canned_reports(self):
        return self.report_dal.get_canned_report_names()

    def cancel_pipeline_instance(self, pipeline_instance_id):
        return self.pipeline_dal.cancel_pipeline_instance(pipeline_instance_id)

    def cancel_action_instance(self, action_instance_id):
        return self.action_dal.cancel_action_instance(action_instance_id)

    def get_work_request_by_action_instance_id(self, action_instance_id):
        return self.action_dal.get_work_request_by_action_instance_id(action_instance_id)



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

adding indexes

Revision ID: 3da597dde134
Revises: 24054f4c1f76
Create Date: 2015-12-17 13:37:27.681085

"""

# revision identifiers, used by Alembic.
revision = '3da597dde134'
down_revision = '24054f4c1f76'
branch_labels = ()
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    if 'sqlite' != op.get_context().dialect.name:
        op.alter_column('actions', 'name',
                        existing_type=sa.VARCHAR(length=100),
                        nullable=False)

    ### commands auto generated by Alembic - please adjust! ###
    op.create_index(op.f('ix_action_instances_action_id'), 'action_instances', ['action_id'], unique=False)
    op.create_index(op.f('ix_action_instances_id'), 'action_instances', ['id'], unique=False)
    op.create_index(op.f('ix_action_instances_pipeline_instance_id'), 'action_instances', ['pipeline_instance_id'], unique=False)
    op.create_index(op.f('ix_action_instances_status_id'), 'action_instances', ['status_id'], unique=False)
    op.create_index(op.f('ix_action_instances_workflow_instance_id'), 'action_instances', ['workflow_instance_id'], unique=False)
    op.create_index(op.f('ix_actions_id'), 'actions', ['id'], unique=False)
    op.create_index(op.f('ix_actions_pipeline_id'), 'actions', ['pipeline_id'], unique=False)
    op.create_index(op.f('ix_actions_workflow_id'), 'actions', ['workflow_id'], unique=False)
    op.create_index(op.f('ix_pipeline_instances_id'), 'pipeline_instances', ['id'], unique=False)
    op.create_index(op.f('ix_pipeline_instances_pipeline_id'), 'pipeline_instances', ['pipeline_id'], unique=False)
    op.create_index(op.f('ix_pipeline_instances_status_id'), 'pipeline_instances', ['status_id'], unique=False)
    op.create_index(op.f('ix_pipeline_parameters_id'), 'pipeline_parameters', ['id'], unique=False)
    op.create_index(op.f('ix_pipeline_parameters_parameter'), 'pipeline_parameters', ['parameter'], unique=False)
    op.create_index(op.f('ix_pipeline_parameters_pipeline_instance_id'), 'pipeline_parameters', ['pipeline_instance_id'], unique=False)
    op.create_index(op.f('ix_pipeline_statistics_id'), 'pipeline_statistics', ['id'], unique=False)
    op.create_index(op.f('ix_pipeline_statistics_pipeline_instance_id'), 'pipeline_statistics', ['pipeline_instance_id'], unique=False)
    op.create_index(op.f('ix_pipeline_statistics_statistics_id'), 'pipeline_statistics', ['statistics_id'], unique=False)
    op.create_index(op.f('ix_pipelines_id'), 'pipelines', ['id'], unique=False)
    op.create_index(op.f('ix_qa_test_histories_action_instance_id'), 'qa_test_histories', ['action_instance_id'], unique=False)
    op.create_index(op.f('ix_qa_test_histories_id'), 'qa_test_histories', ['id'], unique=False)
    op.create_index(op.f('ix_qa_test_histories_pipeline_instance_id'), 'qa_test_histories', ['pipeline_instance_id'], unique=False)
    op.create_index(op.f('ix_qa_test_histories_status_id'), 'qa_test_histories', ['status_id'], unique=False)
    op.create_index(op.f('ix_qa_test_histories_test_id'), 'qa_test_histories', ['test_id'], unique=False)
    op.create_index(op.f('ix_qa_tests_id'), 'qa_tests', ['id'], unique=False)
    op.create_index(op.f('ix_stage_instances_id'), 'stage_instances', ['id'], unique=False)
    op.create_index(op.f('ix_stage_instances_pipeline_instance_id'), 'stage_instances', ['pipeline_instance_id'], unique=False)
    op.create_index(op.f('ix_stage_instances_stage_id'), 'stage_instances', ['stage_id'], unique=False)
    op.create_index(op.f('ix_stage_instances_status_id'), 'stage_instances', ['status_id'], unique=False)
    op.create_index(op.f('ix_stages_id'), 'stages', ['id'], unique=False)
    op.create_index(op.f('ix_stages_pipeline_id'), 'stages', ['pipeline_id'], unique=False)
    op.create_index(op.f('ix_statistics_id'), 'statistics', ['id'], unique=False)
    op.create_index(op.f('ix_statuses_id'), 'statuses', ['id'], unique=False)
    op.create_index(op.f('ix_statuses_type'), 'statuses', ['type'], unique=False)
    op.create_index(op.f('ix_workflow_instances_id'), 'workflow_instances', ['id'], unique=False)
    op.create_index(op.f('ix_workflow_instances_stage_instance_id'), 'workflow_instances', ['stage_instance_id'], unique=False)
    op.create_index(op.f('ix_workflow_instances_status_id'), 'workflow_instances', ['status_id'], unique=False)
    op.create_index(op.f('ix_workflow_instances_workflow_id'), 'workflow_instances', ['workflow_id'], unique=False)
    op.create_index(op.f('ix_workflows_id'), 'workflows', ['id'], unique=False)
    op.create_index(op.f('ix_workflows_stage_id'), 'workflows', ['stage_id'], unique=False)
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f('ix_workflows_stage_id'), table_name='workflows')
    op.drop_index(op.f('ix_workflows_id'), table_name='workflows')
    op.drop_index(op.f('ix_workflow_instances_workflow_id'), table_name='workflow_instances')
    op.drop_index(op.f('ix_workflow_instances_status_id'), table_name='workflow_instances')
    op.drop_index(op.f('ix_workflow_instances_stage_instance_id'), table_name='workflow_instances')
    op.drop_index(op.f('ix_workflow_instances_id'), table_name='workflow_instances')
    op.drop_index(op.f('ix_statuses_type'), table_name='statuses')
    op.drop_index(op.f('ix_statuses_id'), table_name='statuses')
    op.drop_index(op.f('ix_statistics_id'), table_name='statistics')
    op.drop_index(op.f('ix_stages_pipeline_id'), table_name='stages')
    op.drop_index(op.f('ix_stages_id'), table_name='stages')
    op.drop_index(op.f('ix_stage_instances_status_id'), table_name='stage_instances')
    op.drop_index(op.f('ix_stage_instances_stage_id'), table_name='stage_instances')
    op.drop_index(op.f('ix_stage_instances_pipeline_instance_id'), table_name='stage_instances')
    op.drop_index(op.f('ix_stage_instances_id'), table_name='stage_instances')
    op.drop_index(op.f('ix_qa_tests_id'), table_name='qa_tests')
    op.drop_index(op.f('ix_qa_test_histories_test_id'), table_name='qa_test_histories')
    op.drop_index(op.f('ix_qa_test_histories_status_id'), table_name='qa_test_histories')
    op.drop_index(op.f('ix_qa_test_histories_pipeline_instance_id'), table_name='qa_test_histories')
    op.drop_index(op.f('ix_qa_test_histories_id'), table_name='qa_test_histories')
    op.drop_index(op.f('ix_qa_test_histories_action_instance_id'), table_name='qa_test_histories')
    op.drop_index(op.f('ix_pipelines_id'), table_name='pipelines')
    op.drop_index(op.f('ix_pipeline_statistics_statistics_id'), table_name='pipeline_statistics')
    op.drop_index(op.f('ix_pipeline_statistics_pipeline_instance_id'), table_name='pipeline_statistics')
    op.drop_index(op.f('ix_pipeline_statistics_id'), table_name='pipeline_statistics')
    op.drop_index(op.f('ix_pipeline_parameters_pipeline_instance_id'), table_name='pipeline_parameters')
    op.drop_index(op.f('ix_pipeline_parameters_parameter'), table_name='pipeline_parameters')
    op.drop_index(op.f('ix_pipeline_parameters_id'), table_name='pipeline_parameters')
    op.drop_index(op.f('ix_pipeline_instances_status_id'), table_name='pipeline_instances')
    op.drop_index(op.f('ix_pipeline_instances_pipeline_id'), table_name='pipeline_instances')
    op.drop_index(op.f('ix_pipeline_instances_id'), table_name='pipeline_instances')
    op.drop_index(op.f('ix_actions_workflow_id'), table_name='actions')
    op.drop_index(op.f('ix_actions_pipeline_id'), table_name='actions')
    op.drop_index(op.f('ix_actions_id'), table_name='actions')
    if 'sqlite' != op.get_context().dialect.name:
        op.alter_column('actions', 'name',
                   existing_type=sa.VARCHAR(length=100),
                   nullable=True)
    op.drop_index(op.f('ix_action_instances_workflow_instance_id'), table_name='action_instances')
    op.drop_index(op.f('ix_action_instances_status_id'), table_name='action_instances')
    op.drop_index(op.f('ix_action_instances_pipeline_instance_id'), table_name='action_instances')
    op.drop_index(op.f('ix_action_instances_id'), table_name='action_instances')
    op.drop_index(op.f('ix_action_instances_action_id'), table_name='action_instances')
    ### end Alembic commands ###



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
# pylint: disable=too-many-instance-attributes
import datetime
import logging

from rapid.lib.store_service import StoreService
from rapid.lib.constants import StatusTypes, status_type_severity_mapping, StatusConstants
from rapid.workflow.data.dal.status_dal import StatusDal
from rapid.workflow.data.models import Pipeline, PipelineInstance, Stage, StageInstance, Action, \
    ActionInstance, Workflow, WorkflowInstance
logger = logging.getLogger('rapid')


class WorkflowEngine(object):
    def __init__(self, pipeline):
        self.pipeline = pipeline

        self.pipeline_class = Pipeline
        self.stage_class = Stage
        self.workflow_class = Workflow
        self.action_class = Action

        self.stage_mapper = {}
        self.workflow_mapper = {}
        self.action_mapper = {}
        self.status_cache = {}
        self.__loaded__ = False

    def _load_pipeline(self):
        if not self.__loaded__:
            for stage in self._get_stages(self.pipeline):
                self.stage_mapper[stage.id] = stage
                for workflow in self._get_workflows(stage):
                    self.workflow_mapper[workflow.id] = workflow
                    for action in self._get_actions(workflow):
                        self.action_mapper[action.id] = action
            self.__loaded__ = True
        return self.pipeline

    def _get_stages(self, pipeline):
        return pipeline.stages

    def _get_workflows(self, stage):
        return stage.workflows

    def _get_workflows_by_stage_id(self, stage_id):
        return self._get_workflows(self._get_stage(stage_id))

    def _get_actions(self, workflow):
        return workflow.actions

    def _get_actions_by_workflow_id(self, workflow_id):
        return self._get_actions(self._get_workflow(workflow_id))

    def _get_action(self, _id):
        self._load_pipeline()
        return self.action_mapper[_id]

    def _get_workflow(self, _id):
        self._load_pipeline()
        return self.workflow_mapper[_id]

    def _get_stage(self, _id):
        self._load_pipeline()
        return self.stage_mapper[_id]


class InstanceWorkflowEngine(WorkflowEngine):
    __injectables__ = {'status_dal': StatusDal}

    def __init__(self, status_dal, pipeline_instance):
        super(InstanceWorkflowEngine, self).__init__(pipeline_instance)
        self.status_dal = status_dal

        self.pipeline_class = PipelineInstance
        self.stage_class = StageInstance
        self.workflow_class = WorkflowInstance
        self.action_class = ActionInstance
        self.instances_to_add = []

    def _get_stages(self, pipeline):
        return pipeline.stage_instances

    def _get_workflows(self, stage):
        return stage.workflow_instances

    def _get_actions(self, workflow):
        return workflow.action_instances

    def complete_an_action(self, action_instance_id, status_id):
        """
        1. Load all the action_instances
        1a. Set status
        2. If status is "successful"
        2a. If sliced, check all others if done. If last move to 3.
        3. If last in workflow, record date end with workflow, status to highest severity
        4. If successful workflow, check all other workflows. If done, set status and date for stage.
        5. Create new stageInstance if another exists.
        6. If last stage, complete the Pipeline Instance
        :param status_id:
        :param action_instance_id:
        """
        pipeline_id = self.pipeline.id
        StoreService.set_calculating_workflow(pipeline_id)

        try:
            action_instance = self._get_action(action_instance_id)
            self._mark_action_instance_complete(action_instance, status_id)

            action_instances = self._get_actions_by_workflow_id(action_instance.workflow_instance_id)
            if self._can_continue_workflow(action_instances):
                if self._activate_next_action_all_complete(action_instances):
                    self.complete_a_workflow(action_instance.workflow_instance_id, StatusConstants.SUCCESS)
            else:
                for instance in action_instances:
                    if instance.status_id == StatusConstants.NEW:
                        self._mark_action_instance_complete(instance, StatusConstants.CANCELED)
                status = StatusConstants.FAILED if status_id != StatusConstants.CANCELED else StatusConstants.CANCELED

                self.complete_a_workflow(action_instance.workflow_instance_id, status)
        finally:
            StoreService.clear_calculating_workflow(pipeline_id)

    def complete_a_workflow(self, workflow_instance_id, status_id):
        workflow_instance = self._get_workflow(workflow_instance_id)
        self._mark_workflow_instance_complete(workflow_instance, status_id)

        # Are all the other workflows complete?
        all_complete, all_passed, failed_type = self._check_and_verify_statuses(self._get_workflows_by_stage_id(workflow_instance.stage_instance_id))

        if all_complete:
            self.complete_a_stage(workflow_instance.stage_instance_id,
                                  StatusConstants.SUCCESS if all_passed else failed_type.id)

    def complete_a_stage(self, stage_instance_id, status_id):
        stage_instance = self._get_stage(stage_instance_id)
        self._mark_stage_instance_complete(stage_instance, status_id)

        if self._reconciled_stages(self._load_pipeline(), stage_instance):
            all_complete, all_passed, failed_type = self._check_and_verify_statuses(self._get_stages(self.pipeline))

            if all_complete:
                self.complete_a_pipeline(StatusConstants.SUCCESS if all_passed else failed_type.id)

    def complete_a_pipeline(self, status_id):
        pipeline_instance = self._load_pipeline()
        self._mark_pipeline_instance_complete(pipeline_instance, status_id)

    def _reconciled_stages(self, pipeline_instance, stage_instance):
        if StatusConstants.SUCCESS == stage_instance.status_id and len(self._get_stages(pipeline_instance)) < len(pipeline_instance.pipeline.stages):
            _next = False
            for stage in pipeline_instance.pipeline.stages:
                if stage.id == stage_instance.stage_id:
                    _next = True
                elif _next:
                    new_instance = stage.convert_to_instance()
                    new_instance.start_date = datetime.datetime.utcnow()
                    new_instance.status_id = StatusConstants.INPROGRESS
                    new_instance.pipeline_instance_id = pipeline_instance.id
                    for workflow in stage.workflows:
                        w_instance = workflow.convert_to_instance()
                        w_instance.start_date = datetime.datetime.utcnow()
                        self._create_action_instances(pipeline_instance, w_instance, workflow)
                        new_instance.workflow_instances.append(w_instance)
                        self.instances_to_add.append(w_instance)
                    self.instances_to_add.append(new_instance)
                    break
            return False
        if StatusConstants.SUCCESS == stage_instance.status_id:
            found = False
            for s_instance in self._get_stages(pipeline_instance):
                if found:
                    s_instance.start_date = datetime.datetime.utcnow()
                    s_instance.status_id = StatusConstants.INPROGRESS

                    for w_instance in self._get_workflows(s_instance):
                        w_instance.start_date = datetime.datetime.utcnow()
                        w_instance.status_id = StatusConstants.INPROGRESS

                        first_order = -1
                        for action in self._get_actions(w_instance):
                            if first_order < 0 or first_order == action.order:
                                action.status_id = StatusConstants.READY
                                first_order = action.order
                    break
                if s_instance.id == stage_instance.id:
                    found = True
            return found
        return True

    def _create_action_instances(self, pipeline_instance, w_instance, workflow):
        first = True
        for action in workflow.actions:
            slices = action.slices
            if slices > 0:
                for _slice in range(1, slices + 1):
                    a_instance = action.convert_to_instance()
                    a_instance.status_id = StatusConstants.READY if first else StatusConstants.NEW
                    a_instance.pipeline_instance_id = pipeline_instance.id
                    a_instance.slice = "{}/{}".format(_slice, action.slices)
                    w_instance.action_instances.append(a_instance)
                    self.instances_to_add.append(a_instance)
                first = False
            else:
                a_instance = action.convert_to_instance()
                a_instance.status_id = StatusConstants.READY if first else StatusConstants.NEW
                a_instance.pipeline_instance_id = pipeline_instance.id
                a_instance.slice = "1/1"
                first = False
                w_instance.action_instances.append(a_instance)
                self.instances_to_add.append(a_instance)

    def _check_and_verify_statuses(self, objects):
        all_complete = True
        all_passed = True
        failed_type = None

        for instance in objects:
            if instance.status_id <= StatusConstants.INPROGRESS:
                all_complete = False
                break

            status = self._get_status(instance.status_id)
            if status.type == StatusTypes.FAILED or status.type == StatusTypes.CANCELED:
                all_passed = False
                failed_type = status

        return all_complete, all_passed, failed_type

    def _get_status(self, status_id):
        if status_id not in self.status_cache:
            self.status_cache[status_id] = self.status_dal.get_status_by_id(status_id)
        return self.status_cache[status_id]

    def _can_continue_workflow(self, actions):
        """
        :param actions:
        :type actions: list[ActionInstance]
        :return:
        :rtype:
        """
        highest_severity = -1
        sliced_statuses = []
        current_action_id = None
        for instance in actions:
            if current_action_id is None:
                current_action_id = instance.action_id
                sliced_statuses.append(int(instance.status_id))
            elif current_action_id == instance.action_id:
                sliced_statuses.append(int(instance.status_id))
            else:
                current_action_id = instance.action_id

            if instance.status_id >= StatusConstants.INPROGRESS:
                status = self._get_status(instance.status_id)

                if status.type in status_type_severity_mapping:
                    severity = status_type_severity_mapping[status.type]
                    highest_severity = max(severity, highest_severity)
        if StatusConstants.NEW in sliced_statuses \
                or StatusConstants.READY in sliced_statuses \
                or StatusConstants.INPROGRESS in sliced_statuses:
            return True

        return highest_severity < status_type_severity_mapping[StatusTypes.CANCELED]

    def _activate_next_action_all_complete(self, action_instances):
        order_check = None
        all_complete = True
        fail_found = False
        for instance in action_instances:
            if instance.status_id != StatusConstants.NEW:
                if instance.status_id == StatusConstants.INPROGRESS:
                    all_complete = False
                    break  # if you are inprogress, don't bother checking the rest.
                elif instance.status_id == StatusConstants.READY:
                    all_complete = False
                elif instance.status_id > StatusConstants.SUCCESS:
                    fail_found = True
                order_check = instance.order
                continue

            if fail_found:
                break

            if order_check is None or order_check == instance.order:
                instance.status_id = StatusConstants.READY
                order_check = instance.order
                all_complete = False
            if all_complete and order_check < instance.order:
                instance.status_id = StatusConstants.READY
                all_complete = False
                order_check = instance.order
                # break
        return all_complete

    def reset_pipeline(self):
        now = datetime.datetime.utcnow()
        self.pipeline.status_id = StatusConstants.INPROGRESS
        self.pipeline.start_date = now
        self.pipeline.end_date = None

        first_stage = True
        for stage in self._get_stages(self.pipeline):
            self.reset_stage(stage, first_stage, now)
            first_stage = False

    def reset_stage(self, stage, first_stage=False, now=None):
        now = datetime.datetime.utcnow() if now is None else now

        stage.status_id = StatusConstants.READY if first_stage else StatusConstants.NEW
        stage.created_date = now
        stage.start_date = now if first_stage else None
        stage.end_date = None

        for workflow in self._get_workflows(stage):
            self.reset_workflow(workflow, first_stage, now)

    def reset_workflow(self, workflow, first_stage=False, now=None):
        now = datetime.datetime.utcnow() if now is None else now

        workflow.status_id = StatusConstants.INPROGRESS if first_stage else StatusConstants.NEW
        workflow.start_date = now if first_stage else None
        workflow.end_date = None

        actions = self._get_actions(workflow)
        self.reset_action_instances(actions[0], actions, first_stage)

    def get_action_instance_by_id(self, action_id):
        for stage in self._get_stages(self.pipeline):
            for workflow in self._get_workflows(stage):
                for action in self._get_actions(workflow):
                    if action.id == action_id:
                        return action
        return None

    def reset_action(self, action, first_stage=False):
        action.workflow_instance.status_id = StatusConstants.INPROGRESS if first_stage else StatusConstants.NEW
        action.workflow_instance.end_date = None
        action.workflow_instance.stage_instance.status_id = StatusConstants.INPROGRESS if first_stage else StatusConstants.NEW
        action.workflow_instance.stage_instance.end_date = None

        self.reset_action_instances(action, action.workflow_instance.action_instances, first_stage, single_instance=True)

    def reset_action_instances(self, action_instance, action_instances, first_stage=False, single_instance=False):
        first = True
        first_order = None
        action_order = action_instance.order
        for instance in action_instances:
            current_order = instance.order
            if first_order is None:
                first_order = current_order

            if first and first_order != current_order:
                first = False

            if single_instance:
                if current_order < action_order or action_instance.id == instance.id:
                    if action_instance.id == instance.id:
                        if single_instance:
                            instance.workflow_instance.status_id = StatusConstants.INPROGRESS if not first_stage else StatusConstants.NEW
                            instance.workflow_instance.end_date = None
                            instance.workflow_instance.stage_instance.end_date = None
                            instance.workflow_instance.stage_instance.status_id = StatusConstants.INPROGRESS if not first_stage else StatusConstants.NEW
                            self.pipeline.end_date = None
                            self.pipeline.status_id = StatusConstants.INPROGRESS

                    instance.status_id = StatusConstants.NEW if not first else StatusConstants.READY
                    instance.start_date = None
                    instance.assigned_to = None
                    instance.end_date = None
                elif instance.order == current_order:
                    continue
                else:
                    break
            else:
                instance.status_id = StatusConstants.NEW if not (first and first_stage) else StatusConstants.READY
                instance.start_date = None
                instance.assigned_to = None
                instance.end_date = None


    @staticmethod
    def _mark_action_instance_complete(action_instance, status_id):
        action_instance.status_id = status_id
        action_instance.end_date = datetime.datetime.utcnow() if status_id != StatusConstants.CANCELED else action_instance.start_date

    @staticmethod
    def _mark_workflow_instance_complete(workflow_instance, status_id):
        workflow_instance.status_id = status_id
        workflow_instance.end_date = datetime.datetime.utcnow()

    @staticmethod
    def _mark_stage_instance_complete(stage_instance, status_id):
        stage_instance.status_id = status_id
        stage_instance.end_date = datetime.datetime.utcnow() if status_id != StatusConstants.CANCELED else stage_instance.start_date

    @staticmethod
    def _mark_pipeline_instance_complete(pipeline_instance, status_id):
        pipeline_instance.status_id = status_id
        pipeline_instance.end_date = datetime.datetime.utcnow()



"""
Copyright (c) 2018 Michael Bright and Bamboo HR LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


class ConfigurationGenerator(object):
    """
    ConfigurationGenerator is used to generator a configuration file for rapid.
    """
    def generate(self, config_type, config_file=None):
        config = None
        if config_type == 'master':
            from rapid.master.master_configuration import MasterConfiguration
            config = MasterConfiguration()

        elif config_type == 'client':
            from rapid.client.client_configuration import ClientConfiguration
            config = ClientConfiguration()
        self.gather_output(config, config_type, config_file)

    def gather_output(self, config, config_type, config_file):
        if config_file is not None:
            config.parse_config_file(config_file)
        else:
            for key, value in config.__dict__.items():
                if key in ['hostname']:
                    continue

                read_input = None
                try:
                    read_input = raw_input("Enter value for '{}'[{}]? ".format(key, value))
                except NameError:
                    read_input = input(prompt="Enter the value for '{}'[{}]? ".format(key, value))

                if not read_input:
                    read_input = value

                setattr(config, key, read_input)

        print("---------------------------")
        print("### {}.cfg".format(config_type))
        print("---------------------------")
        for section, mappings in config.section_mapping.items():
            print('[{}]'.format(section))
            for key in sorted(mappings.keys()):
                value = getattr(config, key)
                if type(value) == str and "\n" in value:
                    value = value.replace("\n", "\n  ")
                print("{}:{}".format(key, value))
            print('')



"""
Copyright (c) 2015 Michael Bright and Bamboo HR LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging
import glob
import os
import subprocess

from flask import Response
logger = logging.getLogger('rapid')


class LogServer(object):
    def __init__(self, log_dir=None):
        self.log_dir = log_dir

    def _read_log(self, grep):
        if self.log_dir is None:
            yield "The logging directory is not configured."

        try:
            string_grep = "__RCI_{}__".format(grep)

            found_output = False
            gz_files = glob.glob(os.path.join(self.log_dir, "*.gz"))
            log_files = glob.glob(os.path.join(self.log_dir, "*.log"))

            if log_files:
                for check in [{'cmd': 'grep -a {} {}', 'files': ' '.join(log_files)}, {'cmd': 'zgrep -a {} {}', 'files': ' '.join(gz_files)}]:
                    cmd = check['cmd'].format(string_grep, check['files'])
                    process = subprocess.Popen(cmd.split(), stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
                    for line in iter(process.stdout.readline, b''):
                        found_output = True
                        yield line

                    if found_output:
                        break
            else:
                yield ""
        except Exception as exception:  # pylint: disable=broad-except
            logger.exception(exception)
        yield ""

    def configure_application(self, flask_app):
        flask_app.add_url_rule("/log/<path:grep>", 'logging', self.grep_log, methods=['GET'])

        if self.log_dir:
            flask_app.log_dir = self.log_dir

    def grep_log(self, grep):
        return Response(self._read_log(grep), content_type='text/plain')



"""
Copyright (c) 2015 Michael Bright and Bamboo HR LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import json
import logging
import requests

from rapid.lib.constants import EventTypes
from rapid.workflow.events.event_handler import EventHandler
# pylint: disable=broad-except

logger = logging.getLogger('rapid')


class RemoteNotification(object):
    def __init__(self, config):
        self.headers = {}
        self.url = None
        self.payload = None
        self.verify = True
        
        for key in ['headers', 'url', 'payload', 'verify']:
            try:
                setattr(self, key, config[key])
            except Exception:
                pass

    def send(self):
        response = requests.post(self.url, headers=self.headers, json=self.payload, verify=self.verify, timeout=5)
        if response.status_code != 200:
            logger.error("Remote Notification Failed to send to [{}] with status: {}".format(self.url, response.status_code))
        return response


class RemoteNotificationHandler(EventHandler):
    @staticmethod
    def get_event_type():
        """
        :return:
        :rtype: EventTypes
        """
        return EventTypes.RemoteNotification

    def handle_event(self, pipeline_instance, action_instance, event):
        """
        :param pipeline_instance:
        :type pipeline_instance: rapid.workflow.data.models.PipelineInstance
        :param action_instance:
        :type action_instance: rapid.workflow.data.models.ActionInstance
        :param event:
        :type event: rapid.workflow.data.models.PipelineEvent
        :return:
        :rtype:
        """
        if self.passes_conditional(pipeline_instance, action_instance, event.conditional):
            config_dict = {}
            try:
                config_dict = json.loads(event.config)
            except Exception:
                pass
            config = self.prepare(config_dict, pipeline_instance, action_instance)
            notification = RemoteNotification(config)
            response = notification.send()

            if response.status_code >= 400:
                logger.error("Remote Notification was not sent for: {} with status: {}".format(pipeline_instance.id, response.status_code))

    def prepare(self, config, pipeline_instance, action_instance):
        parameters = pipeline_instance.get_parameters_dict()
        config['payload'] = self.prepare_payload(config, pipeline_instance, action_instance, parameters)
        config['headers'] = self.prepare_headers(config, pipeline_instance, action_instance, parameters)
        config['url'] = self.prepare_url(config, pipeline_instance, action_instance, parameters)
        return config

    def prepare_headers(self, config, pipeline_instance, action_instance, parameters):
        headers = None
        try:
            headers = self.__replace_with_parameters(config['headers'], pipeline_instance, action_instance, parameters)
        except Exception:
            pass
        return headers

    def prepare_url(self, config, pipeline_instance, action_instance, parameters):
        new_url = None
        try:
            new_url = config['url']
            new_url = self._translate_string_for_pipeline_instance(pipeline_instance, new_url)
            new_url = self._translate_string_for_action_instance(action_instance, new_url)

            if '{' in new_url and '}' in new_url:
                for key, value in parameters.items():
                    if "{{{}}}".format(key) in new_url:
                        new_url = new_url.replace('{{{}}}'.format(key), value)
        except Exception:
            pass
        return new_url

    def prepare_payload(self, config, pipeline_instance, action_instance, parameters):
        """

        :param config:
        :type config: dict
        :type parameters: dict
        :return:
        :rtype: dict
        """
        payload = None
        try:
            payload = self.__replace_with_parameters(config['payload'], pipeline_instance, action_instance, parameters)
        except Exception:
            pass
        return payload

    def __replace_with_parameters(self, config_dict, pipeline_instance, action_instance, parameters):
        new_dict = config_dict
        try:
            for key, value in config_dict.items():
                key = self._translate_string_for_pipeline_instance(pipeline_instance, key)
                key = self._translate_string_for_action_instance(action_instance, key)

                value = self._translate_string_for_pipeline_instance(pipeline_instance, value)
                value = self._translate_string_for_action_instance(action_instance, value)

                try:
                    if value.startswith('{') and value[-1] == '}':
                        new_dict[key] = parameters[value.replace('{', '').replace('}', '')]
                    elif '{' in value and '}' in value:
                        new_dict[key] = value.replace('{', '').replace('}', '')
                except Exception:
                    pass
        except Exception:
            pass
        return new_dict



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

from .injectable import Injectable


class IOC(object):
    instance = None
    global_registers = {}

    def get_instance_of(self, clzz, *args):
        if clzz is not None:
            if issubclass(clzz, Injectable):
                real_args = []
                for param in clzz.__init__.__code__.co_varnames:
                    if param in clzz.__injectables__:
                        if param in self.global_registers:
                            real_args.append(self.global_registers[param])
                        else:
                            real_args.append(self.get_instance_of(clzz.__injectables__[param]))
                real_args.extend(args)
                return clzz(*real_args)
            return clzz(*args)

    @staticmethod
    def get_instance():
        if IOC.instance is None:
            IOC.instance = IOC()
        return IOC.instance

    @staticmethod
    def get_class_instance(clzz, *args):
        return IOC.get_instance().get_instance_of(clzz, *args)

    @staticmethod
    def register_global(name, value):
        IOC.get_instance().global_registers[name] = value



import logging

from flask import Flask

from rapid.client import ClientCommunicator
from rapid.client.client_configuration import ClientConfiguration
from rapid.lib.utils import UpgradeUtil
from rapid.lib.version import Version

logger = logging.getLogger('rapid')


class ClientUpgrader(object):
    def __init__(self, rapid_config):
        # type: (ClientConfiguration) -> None
        self.client_config = rapid_config

    def upgrade(self, app):
        # type: (Flask) -> None
        communicator = ClientCommunicator(self.client_config.master_uri, self.client_config.quarantine_directory, app, self.client_config.verify_certs)
        self.client_config.is_single_use = True

        info = communicator.register(self.client_config)
        if info['master_version'] != Version.get_version():
            logger.info("Found Version: {}".format(info['master_version']))
            # UpgradeUtil.upgrade_version(info['master_version'], self.client_config)
            UpgradeUtil.upgrade_version(info['master_version'], self.client_config)
        else:
            logger.info("Versions match. No need to upgrade")



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

from abc import ABCMeta, abstractmethod


class QaModule(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def save_results(self, action_instance, session, post_data):
        yield

    @abstractmethod
    def reset_results(self, action_instance_id, session):
        """
        Reset Stacktrace and QaTestHistory entries for action_instance
        :param action_instance_id: int
        :param session:
        :return:
        """
        yield

    @abstractmethod
    def analyze_tests(self, pipeline_instance_id, json):
        """
        Analyze the Test definitions for rapid-* comments
        :param pipeline_instance_id:
        :type pipeline_instance_id: int
        :param json: JSON representation for data
        :type json: dict
        :return: dictionary
        :rtype: dict
        """
        yield

    @abstractmethod
    def get_test_results(self, action_instance_id, filters=None):
        """
        Get test results by filter
        :param action_instance_id:
        :type action_instance_id: int
        :param filters:
        :type filters: string
        :return:
        :rtype: dict
        """
        yield

    @abstractmethod
    def get_qa_testmap_coverage(self, pipeline_instance_id):
        yield


class WorkflowModule(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def start_pipeline_via_repo(self, repo_name, parameters):
        yield

    @abstractmethod
    def get_status_by_name(self, status_name, session=None):
        yield

    @abstractmethod
    def start_pipeline_by_id(self, pipeline_id, json_data=None):
        yield

    @abstractmethod
    def get_pipeline_instance_by_id(self, pipeline_instance_id):
        yield


class CiModule(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_vcs_by_repo_name(self, repo_name):
        yield

    @abstractmethod
    def create_git_commit(self, commit_identifier, vcs_id, pipeline_instance_id=None, additional_parameters=None, session=None):
        yield

    @abstractmethod
    def get_by_identifier(self, commit_identifier):
        yield

    @abstractmethod
    def get_vcs_by_pipeline_id(self, pipeline_id, session=None):
        yield



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

try:
    import simplejson as json
except ImportError:
    import json

import requests
from rapid.lib.version import Version
# pylint: disable=too-many-instance-attributes


class Client(object):

    def __init__(self, ip_address, port, grains, grain_restrict, api_key=None, is_ssl=False, hostname=None, time_elapse=0.5):
        self.ip_address = ip_address
        self.is_ssl = is_ssl
        self.port = port
        self.grains = grains.split(';') if grains is not None else None
        self.grain_restrict = grain_restrict
        self.sleep = False
        self.api_key = api_key
        self.hostname = hostname
        self.time_elapse = ((time_elapse / 1000) + 0.5) if time_elapse > 1 else time_elapse

    def __getstate__(self):
        state = self.__dict__.copy()
        return state

    def get_headers(self):
        return {'X-Rapidci-Version': Version.get_version(), 'Content-Type': 'application/json', 'X-Rapidci-Api-Key': self.api_key}

    def can_handle(self, grain):
        if self.sleep:
            return False

        if grain is not None and self.grains is not None:
            tmp_grains = grain.split(';')
            number_matched = 0
            for _grain in tmp_grains:
                if _grain not in self.grains:
                    return False
                number_matched += 1

            if self.grain_restrict and number_matched != len(self.grains):
                return False

            return True

        if self.grain_restrict:
            return False
        
        return True

    def get_availability_uri(self):
        return "{}://{}/work/request".format(self._get_prefix(), self.get_uri())

    def get_work_uri(self):
        return "{}://{}/work/execute".format(self._get_prefix(), self.get_uri())

    def get_cancel_uri(self, action_instance_id):
        return "{}://{}/work/cancel/{}".format(self._get_prefix(), self.get_uri(), action_instance_id)

    def get_status_uri(self):
        return "{}://{}/status".format(self._get_prefix(), self.get_uri())

    def send_work(self, work_request, verify_certs=True):
        return requests.post(self.get_work_uri(), json=work_request.__dict__, headers=self.get_headers(), verify=verify_certs, timeout=4)

    def cancel_work(self, action_instance_id, verify_certs=True):
        return requests.post(self.get_cancel_uri(action_instance_id), headers=self.get_headers(), verify=verify_certs, timeout=4)

    def get_uri(self):
        return "{}:{}".format(self.ip_address, self.port)

    def _get_prefix(self):
        return 'https' if self.is_ssl else 'http'



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
# pylint: disable=broad-except,too-many-public-methods
from rapid.release.release_service import ReleaseService

try:
    import simplejson as json
except ImportError:
    import json

import logging
from werkzeug.exceptions import BadRequestKeyError
from sqlalchemy.orm import joinedload
from sqlalchemy import desc, asc
from flask import Response, request

from rapid.lib import json_response, api_key_required, get_declarative_base, get_db_session
from rapid.lib.constants import ModuleConstants
from rapid.lib.utils import ORMUtil
from rapid.lib.store_service import StoreService
from rapid.lib.work_request import WorkRequestEncoder
from rapid.lib.framework.injectable import Injectable
from rapid.master.data.database.dal import get_dal
from rapid.workflow.action_instances_service import ActionInstanceService
from rapid.workflow.queue_service import QueueService
from rapid.workflow.workflow_service import WorkflowService

logger = logging.getLogger("rapid")


class APIRouter(Injectable):
    __injectables__ = {'action_instance_service': ActionInstanceService,
                       'queue_service': QueueService,
                       ModuleConstants.QA_MODULE: None,
                       'workflow_service': WorkflowService,
                       'release_service': ReleaseService}
    classes, models, table_names = None, None, None
    class_map = {}

    def __init__(self, action_instance_service, queue_service, qa_module, workflow_service, release_service):
        """

        :param action_instance_service:
        :type action_instance_service: ActionInstanceService
        :param queue_service:
        :type queue_service:
        :param qa_module:
        :type qa_module: rapid.lib.modules.modules.QaModule
        :param workflow_service:
        :type workflow_service: WorkflowService
        :type release_service: ReleaseService
        """
        self.app = None
        self.action_instance_service = action_instance_service
        self.queue_service = queue_service
        self.qa_module = qa_module
        self.workflow_service = workflow_service
        self.release_service = release_service

    def register_url_rules(self, flask_app):
        flask_app.add_url_rule('/api/<path:endpoint>', 'api_list', api_key_required(self.list), methods=['GET', 'POST'])
        flask_app.add_url_rule('/api/<path:endpoint>/metadata', 'api_meta', api_key_required(self.metadata), methods=['GET'])
        flask_app.add_url_rule('/api/metadata', 'all_api_meta', api_key_required(self.metadata), methods=['GET'])
        flask_app.add_url_rule('/api/bulk_create', 'bulk_create', api_key_required(self.bulk_create), methods=['PUT'])
        flask_app.add_url_rule('/api/<path:endpoint>', 'create_object', api_key_required(self.create), methods=['PUT'])
        flask_app.add_url_rule('/api/<path:endpoint>/<int:_id>', 'edit_object', api_key_required(self.edit_object), methods=['POST'])
        flask_app.add_url_rule('/api/<path:endpoint>/<int:_id>', 'delete_object', api_key_required(self.delete_object), methods=['DELETE'])
        flask_app.add_url_rule('/api/<path:endpoint>/<int:_id>', 'list_object', api_key_required(self.list_single_object), methods=['GET'])
        flask_app.add_url_rule('/api/action_instances/<int:_id>/done', 'finish_action_instance', api_key_required(self.finish_action_instance), methods=['POST'])
        flask_app.add_url_rule('/api/action_instances/<int:_id>/callback', 'callback_action_instance', api_key_required(self.callback_action_instance), methods=['POST'])
        flask_app.add_url_rule('/api/action_instances/<int:_id>/reset', 'reset_action_instance', api_key_required(self.reset_action_instance), methods=['POST'])
        flask_app.add_url_rule('/api/action_instances/<int:_id>/results', 'action_instance_results', api_key_required(self.action_instance_results), methods=['GET'])
        flask_app.add_url_rule('/api/action_instances/<int:action_instance_id>/work_request', 'action_instance_work_request', api_key_required(self.action_instance_work_request), methods=['GET']),

        flask_app.add_url_rule('/api/action_instances/<int:action_instance_id>/is_completing', 'action_instance_is_completing', api_key_required(self.action_instance_is_completing), methods=['GET'])
        flask_app.add_url_rule('/api/action_instances/<int:action_instance_id>/clear_completing', 'action_instance_clear_completing', api_key_required(self.action_instance_clear_completing), methods=['GET'])

        flask_app.add_url_rule('/api/pipeline_instances/<int:_id>/failure_count', 'failure_count_instance', api_key_required(self.failure_count_instance))
        flask_app.add_url_rule('/api/queue', 'get_queue', api_key_required(self.get_queue), methods=['GET'])
        flask_app.add_url_rule('/api/pipeline_instances/<int:_id>/reset', 'reset_pipeline_instance', api_key_required(self.reset_pipeline_instance), methods=['POST', 'GET'])

        flask_app.add_url_rule("/api/reports/canned/<path:report_name>", 'canned_report', api_key_required(self.canned_report), methods=['GET'])
        flask_app.add_url_rule("/api/reports/canned/list", 'list_canned_report', api_key_required(self.list_canned_reports), methods=['GET'])

        flask_app.add_url_rule("/api/pipeline_instances/<int:pipeline_instance_id>/cancel", "cancel_pipeline_instance", api_key_required(self.cancel_pipeline_instance), methods=['POST'])
        flask_app.add_url_rule("/api/action_instances/<int:action_instance_id>/cancel", "cancel_action_instance", api_key_required(self.cancel_action_instance), methods=['POST'])
        flask_app.add_url_rule("/api/pipeline_instances/<int:pipeline_instance_id>/print", "print_pipeline_instance", api_key_required(self.print_pipeline_instance), methods=['GET'])

        self.app = flask_app

    def _get_args(self):
        try:
            if request.content_type == 'application/json':
                return request.get_json()
        except Exception:
            pass
        return request.args

    def reset_pipeline_instance(self, _id):
        response = self.action_instance_service.reset_pipeline_instance(_id)
        return Response(json.dumps({'message': ('Success!' if response else 'Failure!')}), content_type='application/json', status=(200 if response else 500))

    def get_queue(self):
        return Response(json.dumps(self.queue_service.get_current_work(), cls=WorkRequestEncoder), content_type='application/json')

    def failure_count_instance(self, _id):
        pass

    def action_instance_results(self, _id):
        _filter = None
        try:
            _filter = self._get_args()['filter']
        except (KeyError, TypeError):
            pass

        return Response(json.dumps(self.qa_module.get_test_results(_id, _filter)), content_type='application/json')

    def list_single_object(self, endpoint, _id):
        if self._is_valid(endpoint):
            for session in get_db_session():
                clazz = self.class_map[endpoint]
                query = session.query(clazz).filter(clazz.id == _id)
                allowed_fields = self._get_additional_fields(clazz)
                instance = query.one()
                return Response(json.dumps(instance.serialize(allowed_children=allowed_fields)), content_type='application/json')
        return Response("Not Valid", status=404)

    def list(self, endpoint):
        if self._is_valid(endpoint):
            for session in get_db_session():
                clazz = self._get_clazz(endpoint)
                query = self._get_query(session, clazz)
                query = self._set_filter(clazz, query)
                query = self._set_orderby_direction(query, clazz)
                query = self._set_joins(query)
                query = self._set_limit(query)
                fields = self._get_additional_fields(clazz)
                results = []
                for result in query.all():
                    results.append(result.serialize(fields))
                return Response(json.dumps(results), content_type='application/json')
        else:
            return Response(status=404)

    @json_response()
    def cancel_pipeline_instance(self, pipeline_instance_id):
        return self.workflow_service.cancel_pipeline_instance(pipeline_instance_id)

    @json_response()
    def cancel_action_instance(self, action_instance_id):
        return self.workflow_service.cancel_action_instance(action_instance_id)

    @json_response()
    def action_instance_work_request(self, action_instance_id):
        return self.workflow_service.get_work_request_by_action_instance_id(action_instance_id)

    def _get_clazz(self, endpoint):
        return self.class_map[endpoint]

    @staticmethod
    def _get_query(session, clazz):
        return session.query(clazz)

    def _set_filter(self, clazz, query):
        _filter = self._get_args()['filter'] if 'filter' in self._get_args() else None
        return ORMUtil.get_filtered_query(query, _filter, clazz)

    def _set_joins(self, query):
        if 'joins' in self._get_args():
            if 'orderby' not in self._get_args():
                raise Exception("You can't join with out an orderby")
            current_attribute = None
            queries = []
            for join in self._get_args()['joins'].split(';'):
                split = join.split('=')
                attribute = split[0]

                if attribute.startswith('_'):  # If attribute starts with '_' then it is a root attribute, start from here.
                    current_attribute = None
                    attribute = attribute.split('_', 1)[1]

                if current_attribute is None:
                    current_attribute = attribute
                else:
                    current_attribute = "{}.{}".format(current_attribute, attribute)

                queries.append(joinedload(current_attribute))

                if len(split) > 1:
                    for field in split[1].split(','):
                        queries.append(joinedload("{}.{}".format(current_attribute, field)))
            query = query.options(queries)
        return query

    def _set_orderby_direction(self, query, clazz):
        if 'orderby' in self._get_args():
            _map = {'desc': desc, 'asc': asc}
            for orderby in self._get_args()['orderby'].split(','):
                try:
                    column, order = orderby.split(':')
                    if hasattr(clazz, column):
                        column = getattr(clazz, column)
                        __import__(clazz.__module__, fromlist=[clazz.__class__])
                    query = query.order_by(_map[order](column))
                except Exception:
                    import traceback
                    traceback.print_exc()
        return query

    def _get_additional_fields(self, clazz, fields=None):
        fields = {clazz.__tablename__: []} if fields is None else fields
        try:
            for tmp_string in self._get_args()['fields'].split(';'):
                relation_split = tmp_string.split('=')
                if len(relation_split) == 1:
                    for field in relation_split[0].split(','):
                        fields[clazz.__tablename__].append(field)
                else:
                    table_name = relation_split[0]
                    for field in relation_split[1].split(','):
                        try:
                            fields[table_name].append(field)
                        except (AttributeError, TypeError, KeyError):
                            fields[table_name] = []
                            fields[table_name].append(field)
        except Exception:
            pass
        return fields

    def _set_limit(self, query):
        limit = 100

        try:
            limit = int(self._get_args()['limit'])
        except (AttributeError, TypeError, ValueError, BadRequestKeyError, KeyError):
            pass
        return query.limit(limit)

    def bulk_create(self):
        result = {}
        for (endpoint, objects) in request.get_json().items():
            if self._is_valid(endpoint):
                clazz = self.class_map[endpoint]
                result[endpoint] = []
                for _object in objects:
                    result[endpoint].append(self._create_from_request(clazz, _object))
        return Response(json.dumps(result), content_type="application/json")

    def create(self, endpoint):
        if self._is_valid(endpoint):
            clazz = self.class_map[endpoint]
            return Response(json.dumps(self._create_from_request(clazz, request.get_json())), content_type="application/json")
        return Response(status=404)

    def edit_object(self, endpoint, _id):
        if self._is_valid(endpoint):
            clazz = self.class_map[endpoint]
            for session in get_db_session():
                dal = self._retrieve_dal(clazz)
                instance = dal.edit_object(session, clazz, _id, request.json)
                return Response(json.dumps(instance.serialize()), content_type='application/json')
        return Response(status=404)

    def delete_object(self, endpoint, _id):
        if self._is_valid(endpoint):
            clazz = self.class_map[endpoint]
            for session in get_db_session():
                dal = self._retrieve_dal(clazz)
                instance = dal.delete_object(session, clazz, _id)
                return Response(json.dumps(instance.serialize()), content_type='application/json')
        return Response(status=404)

    def reset_action_instance(self, _id):
        try:
            if self.action_instance_service.reset_action_instance(_id, True):
                return Response(json.dumps({"message": "Action instance reset"}), content_type='application/json')
            return Response(json.dumps({"message": "Unable to reset instance"}), content_type='application/json', status=505)
        except Exception as exception:
            logger.error(exception)

        return Response(json.dumps({"message": "Something was wrong"}), content_type='application/json', status=500)

    def callback_action_instance(self, _id):
        return Response(json.dumps(self.action_instance_service.callback(_id, request.get_json())), content_type='application/json')

    def finish_action_instance(self, _id):
        try:
            return Response(json.dumps(self.action_instance_service.finish_action_instance(_id, request.get_json())), content_type='application/json')
        except Exception as exception:
            logger.error(exception)
            return Response("Something went wrong!", status=500)

    def metadata(self, endpoint=None):
        if self._is_valid(endpoint):
            clazz = self.class_map[endpoint]
            return Response(json.dumps(self._explode_class(clazz)), content_type="application/json")
        
        ret_map = {}
        for key, clazz in self.class_map.items():
            ret_map[key] = self._explode_class(clazz)
            ret_map[key]['url'] = "/api/{}/metadata".format(key)
        return Response(json.dumps(ret_map), content_type="application/json")

    @json_response()
    def canned_report(self, report_name):
        return self.workflow_service.get_custom_report(report_name)

    @json_response()
    def list_canned_reports(self):
        return self.workflow_service.list_canned_reports()

    @json_response()
    def action_instance_is_completing(self, action_instance_id):
        return {"status": StoreService.is_completing(action_instance_id)}

    @json_response()
    def action_instance_clear_completing(self, action_instance_id):
        return {"status": StoreService.clear_completing(action_instance_id)}

    @json_response()
    def print_pipeline_instance(self, pipeline_instance_id):
        pipeline_instance = self.workflow_service.get_pipeline_instance_by_id(pipeline_instance_id)
        response = ""
        response = "pipeline: {}, Status: {}, Start Date: {}, End Date:{}\n".format(pipeline_instance.id, pipeline_instance.status_id, pipeline_instance.start_date, pipeline_instance.end_date)
        for stage_instance in pipeline_instance.stage_instances:
            response += "stage: {}, Status: {}, Start Date: {}, End Date: {}\n".format(stage_instance.id, stage_instance.status_id, stage_instance.start_date, stage_instance.end_date)
            for workflow_instance in stage_instance.workflow_instances:
                response += "  workflow: {}, Status: {}, Start Date: {}, End Date: {}\n".format(workflow_instance.id, workflow_instance.status_id, workflow_instance.start_date, workflow_instance.end_date)
                for action_instance in workflow_instance.action_instances:
                    response += "    action: {}, Status: {}, Start Date: {}, End Date: {}, Order: {}, Slice: {}\n".format(action_instance.id, action_instance.status_id, action_instance.start_date, action_instance.end_date, action_instance.order, action_instance.slice)
        return {"message": response}

    def _explode_class(self, clazz):
        columns = {}
        for column in clazz.__table__._columns:  # pylint: disable=protected-access
            columns[column.name] = column.type.__class__.__name__

        for field in clazz().__relationships__():
            columns[field] = "__ON_REQUEST__"

        return columns

    def _is_valid(self, endpoint):
        return endpoint in self._get_models()

    def _get_models(self):
        if self.models is None:
            self._load_objects()

        return self.models

    def _load_objects(self):
        self.classes = []
        self.models = []
        self.table_names = []

        _base = get_declarative_base()
        for clazz in list(_base._decl_class_registry.values()):  # pylint: disable=protected-access
            try:
                self.table_names.append(clazz.__tablename__)
                self.classes.append(clazz)
            except Exception:
                pass
        for table in _base.metadata.tables.items():
            if table[0] in self.table_names:
                clazz = self.classes[self.table_names.index(table[0])]
                name = clazz.__tablename__.lower()
                self.models.append(name)
                self.class_map[name] = clazz

    def _retrieve_dal(self, clazz):
        dal = get_dal(clazz)
        if dal is None:
            raise Exception("Object not found.")
        return dal

    def _create_from_request(self, clazz, _json):
        for session in get_db_session():
            dal = self._retrieve_dal(clazz)
            return dal.create_object(session, clazz, _json)

    @staticmethod
    def _convert_to_dict(row):
        ret_dict = {}
        for column in row.__table__.columns:
            ret_dict[column.name] = str(getattr(row, column.name))
        return ret_dict



"""
 Copyright (c) 2015 Michael Bright and Bamboo HR LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

import argparse
import logging

from .configuration_generator import ConfigurationGenerator

parser = argparse.ArgumentParser(description='Rapid Framework client main control')
parser.add_argument('-f', '--config', dest='config_file', help="config file path")
parser.add_argument('-p', '--port', dest='port', help='Port for the master to listen on')
parser.add_argument('-m', '--master', dest='mode_master', help='Master mode is default', action='store_true')
parser.add_argument('-c', '--client', dest='mode_client', help='Client mode', action='store_true')
parser.add_argument('-l', '--logging', dest='mode_logging', help='Logging mode', action='store_true')
parser.add_argument('-d', '--log_dir', dest='log_dir', help='Logging directory')
parser.add_argument('-q', '--qa_dir', dest="qa_dir", help="QA Dir")
parser.add_argument('-w', '--waitress', dest='waitress', action='store_true', help='Run under waitress')
parser.add_argument('-r', '--run', dest='run', help='Run action_instance_id')
parser.add_argument('-u', '--upgrade', dest='upgrade', action='store_true', help="Upgrade to master's version")
parser.add_argument('--static_file_dir', dest='static_file_dir', help='Static File directory, default is current')
parser.add_argument('--static_basic_auth', dest='basic_auth', help="Basic Auth for static file directory")
parser.add_argument('--downgrade', dest='db_downgrade', help="Downgrade db for alembic")
parser.add_argument('--create_db', action='store_true', dest='createdb', help="Create initial db")
parser.add_argument('--create_migration', dest='migrate', help="Create Migration for alembic")
parser.add_argument('--generate-config', dest='generate_config', help='Generate a default configuration', choices=['master', 'client'])
args = parser.parse_args()

if args.mode_client or args.run or args.upgrade:
    from .client import app, configure_application
elif args.mode_logging:
    from .log_server import app, configure_application
elif args.qa_dir:
    from .testmapper import process_directory
    process_directory('', args.qa_dir, True)
    import sys
    sys.exit(0)
elif args.generate_config:
    import sys
    ConfigurationGenerator().generate(args.generate_config, args.config_file)
    sys.exit(0)
else:
    from .master import app, configure_application

logger = logging.getLogger("rapid")


def setup():
    logger.info("Setting up the application.")
    configure_application(app, args)


if not args.migrate:
    setup()


def main():
    if args.createdb:
        with app.app_context():
            app.db.create_all()  # pylint: disable=no-member
    if args.migrate:
        from .master import create_migration_script
        configure_application(app, args, manual_db_upgrade=True)
        create_migration_script(app, args.migrate)
    elif args.db_downgrade:
        print("Downgraded.")
    elif args.waitress:
        try:
            from waitress import serve
            serve(app, port=(int(args.port or app.rapid_config.port)))  # pylint: disable=no-member
        except ImportError:
            print("Failed to start up the server.")
    elif args.run:
        from rapid.client import run_action_instance
        run_action_instance(args.run)
    elif args.upgrade:
        from rapid.client import upgrade_rapid
        upgrade_rapid()
    else:
        app.run('0.0.0.0', port=int(args.port or app.rapid_config.port))  # pylint: disable=no-member


if __name__ == '__main__':
    main()



